from ..hookspecs import hookimpl
import re

DISCOVER_ALLOW = 'discover-allow'

_re = {}

@hookimpl
def canonicalize_url(config, from_url, to_url, to_url_depth):
    if not DISCOVER_ALLOW in config or not config[DISCOVER_ALLOW]:
        return

    allows = config[DISCOVER_ALLOW]

    desired_depth = None
    permitted = False

    for allow in allows:
        from_compiled = None
        to_compiled = None

        if 'from-url-regex' in allow:
            from_re = allow['from-url-regex']
            if from_re in _re:
                from_compiled = _re[from_re]
            else:
                from_compiled = re.compile(from_re)
                _re[from_re] = from_compiled

        if 'to-url-regex' in allow:
            to_re = allow['to-url-regex']
            if to_re in _re:
                to_compiled = _re[to_re]
            else:
                to_compiled = re.compile(to_re)
                _re[to_re] = to_compiled

        if (not from_compiled or from_compiled.search(from_url)) and (not to_compiled or to_compiled.search(to_url)):
            permitted = True

            if 'depth' in allow:
                depth = int(allow['depth'])

                if desired_depth == None or desired_depth > depth:
                    desired_depth = depth

    if not permitted:
        return False

    # Allowed, use default depth
    if desired_depth == None:
        return True

    return (to_url, desired_depth)

@hookimpl
def config_schema():
    from .. import ConfigSchema
    return ConfigSchema(
        schema = {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'from-url-regex': {
                        'type': 'string',
                    },
                    'to-url-regex': {
                        'type': 'string',
                    },
                    'depth': {
                        'type': 'integer',
                    },
                },
            }
        },
        uischema = {
            "type": "Control",
            "scope": '#/properties/{}'.format(DISCOVER_ALLOW),
            'label': "Only crawl URLs matching these regexes"
        },
        key = DISCOVER_ALLOW,
        sort = 2900,
        group = 'Links',
    )

@hookimpl
def config_default_value():
    return []



DISCOVER_MISSING_DICTIONARIES = 'discover-missing-dictionaries'
RECOMPRESS_ALL = 'recompress-all'
RECOMPRESS_HOST = 'recompress-host'
RECOMPRESS_HASH = 'recompress-hash'
SEED_CRAWL = 'seed-crawl'
TRAIN_DICTIONARY = 'train-dictionary'



from ..hookspecs import hookimpl
from urllib.parse import urlparse
import re

CANONICALIZE_SHOPIFY_PRODUCTS = 'canonicalize-shopify-products'

shopify_product_re = re.compile('^https://([^/]+)/collections/[^/]+/products/([^?#]+)(.*)$')

@hookimpl
def canonicalize_url(config, from_url, to_url, to_url_depth):
    if not CANONICALIZE_SHOPIFY_PRODUCTS in config or not config[CANONICALIZE_SHOPIFY_PRODUCTS]:
        return

    m = shopify_product_re.search(to_url)

    if m:
        return 'https://{}/products/{}{}'.format(m.group(1), m.group(2), m.group(3))

@hookimpl
def config_schema():
    from .. import ConfigSchema
    return ConfigSchema(
        schema = {
          'type': 'boolean',
          'title': 'Canonicalize Shopify product URLs',
        },
        uischema = {
            "type": "Control",
            "scope": '#/properties/{}'.format(CANONICALIZE_SHOPIFY_PRODUCTS)
        },
        key = CANONICALIZE_SHOPIFY_PRODUCTS,
        group = 'Links',
        sort = 10
    )

@hookimpl
def config_default_value():
    return True



from ..hookspecs import hookimpl

MAX_PAGES = 'max-pages'

@hookimpl
def canonicalize_url(conn, config, job_id):
    if MAX_PAGES in config:
        max_pages = config[MAX_PAGES]

        fetched, = conn.execute('SELECT COALESCE(SUM(fetched), 0) FROM dss_job_stats WHERE job_id = ?', [job_id]).fetchone()
        if fetched > max_pages:
            return 'max-pages {} is less than fetched pages ({})'.format(max_pages, fetched)

@hookimpl
def config_schema():
    from .. import ConfigSchema
    return ConfigSchema(
        schema = {
          'type': 'number',
        },
        uischema = {
            "type": "Control",
            "scope": '#/properties/{}'.format(MAX_PAGES)
        },
        key = MAX_PAGES,
        group = 'Links',
        sort = 110,
    )

@hookimpl
def config_default_value():
    return 1000



from datasette_scraper.config import ensure_wal_mode
import sqlite3
import json
import types
from more_itertools import batched
from selectolax.parser import HTMLParser
from urllib.parse import urlparse

_last_html = None
_last_html_parser = None

def get_html_parser(response):
    global _last_html
    global _last_html_parser

    text = response['text']
    if text == _last_html:
        return _last_html_parser

    _last_html_parser = HTMLParser(text)
    _last_html = text
    return _last_html_parser


def get_crawl_config_for_job_id(conn, job_id):
    res = conn.execute('SELECT dss_crawl.config FROM dss_crawl JOIN dss_job ON dss_job.crawl_id = dss_crawl.id WHERE dss_job.id = ?', [job_id])
    config, = res.fetchone()
    config = json.loads(config)
    return config

def add_crawl_queue_items(conn, job_id, urls):
    enqueued = 0

    batch_size = 100
    # Try to efficiently prune URLs we've already queued 
    all_urls = [url[0] for url in urls]
    batched_urls = batched(all_urls, batch_size)

    already_queued = {}
    with conn:
        conn.execute('BEGIN TRANSACTION')
        for batch in batched_urls:
            cur = conn.execute(
                'SELECT url FROM dss_crawl_queue WHERE job_id = {} AND url IN ({})'.format(job_id, ','.join(['?'] * len(batch))),
                batch
            )
            cur.arraysize = 100
            rv = cur.fetchmany()
            for (fetched_url, ) in rv:
                already_queued[fetched_url] = True

        all_urls = [url for url in all_urls if not url in already_queued]
        batched_urls = batched(all_urls, batch_size)
        for batch in batched_urls:
            cur = conn.execute(
                'SELECT url FROM dss_crawl_queue_history WHERE job_id = {} AND url IN ({})'.format(job_id, ','.join(['?'] * len(batch))),
                batch
            )
            cur.arraysize = 100
            rv = cur.fetchmany()
            for (fetched_url, ) in rv:
                already_queued[fetched_url] = True

    urls = [url for url in urls if not url[0] in already_queued]

    for (new_url, new_depth) in urls:
        enqueued += add_crawl_queue_item(conn, job_id, new_url, new_depth)

    return enqueued

def add_crawl_queue_item(conn, job_id, url, depth):
    with conn:
        parsed = urlparse(url)
        host = parsed.hostname

        #print('insert dss_host_rate_limit host={}'.format(host))
        conn.execute('BEGIN TRANSACTION')
        conn.execute('INSERT INTO dss_host_rate_limit(host) SELECT ? WHERE NOT EXISTS(SELECT * FROM dss_host_rate_limit WHERE host = ?)', [host, host])


        cur = conn.execute('INSERT INTO dss_crawl_queue(job_id, host, url, depth) SELECT ?, ?, ?, ? WHERE NOT EXISTS(SELECT * FROM dss_crawl_queue WHERE job_id = ? AND url = ?) AND NOT EXISTS(SELECT * FROM dss_crawl_queue_history WHERE job_id = ? AND url = ?)', [job_id, host, url, depth, job_id, url, job_id, url])

        return cur.rowcount


def reject_crawl_queue_item(conn, id, reason):
    with conn:
        conn.execute('BEGIN TRANSACTION')
        conn.execute("INSERT INTO dss_crawl_queue_history(job_id, host, url, depth, processed_at, fetched_fresh, skipped_reason) SELECT job_id, host, url, depth, strftime('%Y-%m-%d %H:%M:%f'), 0, ? FROM dss_crawl_queue WHERE id = ?", [reason, id])
        conn.execute("DELETE FROM dss_crawl_queue WHERE id = ?", [id])

def finish_crawl_queue_item(conn, id, response, fresh, fetch_duration):
    content_type = 'application/octet-stream'
    status_code = response['status_code']

    # just an approximation -- not, e.g., the Content-Length header
    size = len(response['text'])

    for header in response['headers']:
        if header[0] == 'content-type':
            content_type = header[1].split(';')[0]

    with conn:
        conn.execute('BEGIN TRANSACTION')
        conn.execute("INSERT INTO dss_crawl_queue_history(job_id, host, url, depth, processed_at, fetched_fresh, status_code, content_type, size, duration, request_hash) SELECT job_id, host, url, depth, strftime('%Y-%m-%d %H:%M:%f'), ?, ?, ?, ?, ?, ? FROM dss_crawl_queue WHERE id = ?", [fresh, status_code, content_type, size, fetch_duration, response['_request_hash'] if '_request_hash' in response else None, id])
        conn.execute("DELETE FROM dss_crawl_queue WHERE id = ?", [id])

def check_for_job_complete(conn, job_id):
    more_to_do, = conn.execute("SELECT EXISTS(SELECT * FROM dss_crawl_queue WHERE job_id = ?) AND EXISTS(SELECT * FROM dss_job WHERE id = ? AND status != 'cancelled' AND status != 'done')", [job_id, job_id]).fetchone()
    if not more_to_do:
        with conn:
            conn.execute('BEGIN TRANSACTION')
            conn.execute("UPDATE dss_job SET status = 'done', finished_at = strftime('%Y-%m-%d %H:%M:%f') WHERE id = ? AND finished_at IS NULL", [job_id])
            conn.execute("DELETE FROM dss_crawl_queue WHERE job_id = ?", [job_id])

def lazy_connection_factory(db_map):
    conns = {}

    def get_db(name):
        if not name in db_map:
            raise Exception('unknown database name: {}'.format(name))

        if name in conns:
            return conns[name]

        conn = sqlite3.connect(db_map[name])
        conn.isolation_level = None
        ensure_wal_mode(conn)

        # See https://www.sqlite.org/pragma.html#pragma_synchronous; this is much faster,
        # at the expense of durability in the event of an unplanned shutdown.
        conn.execute('pragma synchronous = normal;')
        conns[name] = conn
        return conn

    return get_db

def lazy_connection_factory_with_default(factory, default):
    def get_db(name):
        if not name:
            return factory(default)

        return factory(name)

    return get_db

def module_from_path(path, name):
    # Stolen from https://github.com/simonw/datasette/blob/013496862f4d4b441ab61255242b838b24287607/datasette/utils/__init__.py#L741
    # Adapted from http://sayspy.blogspot.com/2011/07/how-to-import-module-from-just-file.html
    mod = types.ModuleType(name)
    mod.__file__ = path
    with open(path, "r") as file:
        code = compile(file.read(), path, "exec", dont_inherit=True)
    exec(code, mod.__dict__)
    return mod



import re
from ..hookspecs import hookimpl
from ..utils import get_html_parser
from urllib.parse import urljoin

EXTRACT_LINKS = 'extract-links'

_re = {}

@hookimpl
def extract_from_response(config, url, response):
    if not EXTRACT_LINKS in config:
        return {}

    rv = {}

    for options in config[EXTRACT_LINKS]:
        if 'url-regex' in options:
            regex = options['url-regex']
            if regex in _re:
                compiled = _re[regex]
            else:
                compiled = re.compile(regex)
                _re[regex] = compiled

            if not compiled.search(url):
                continue

        dbname = options['database']
        tablename = options['table']

        if dbname in rv:
            database = rv[dbname]
        else:
            database = rv[dbname] = {}

        if tablename in database:
            table = database[tablename]
        else:
            table = database[tablename] = []

        table.append({'from@': url, '__delete': True})
        # Get all the links
        parsed = get_html_parser(response)
        for a in parsed.css('a'):
            attrs = a.attributes
            dofollow = 1

            if 'rel' in attrs and attrs['rel'] and 'nofollow' in attrs['rel']:
                dofollow = 0


            if 'href' in attrs:
                table.append({'from@': url, 'to': urljoin(url, attrs['href']), 'text': a.text().strip(), 'dofollow': dofollow})

    return rv

@hookimpl
def config_schema():
    from .. import ConfigSchema
    return ConfigSchema(
        schema = {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'url-regex': {
                        'type': 'string',
                    },
                    'database': {
                        'type': 'string',
                    },
                    'table': {
                        'type': 'string',
                    },
                },
                'required': ['database', 'table']
            }
        },
        uischema = {
            "type": "Control",
            "scope": '#/properties/{}'.format(EXTRACT_LINKS),
            'label': 'Extract link graph'
        },
        key = EXTRACT_LINKS,
        group = 'Extracting',
    )

@hookimpl
def config_default_value():
    return []



from pluggy import HookimplMarker
from pluggy import HookspecMarker

hookspec = HookspecMarker("datasette_scraper")
hookimpl = HookimplMarker("datasette_scraper")

@hookspec
def config_schema():
    """Returns a JSON-schema of what a valid value looks like."""

@hookspec
def config_default_value():
    """Returns a reasonable default value that conforms to config_schema."""

@hookspec
def get_seed_urls(conn, config, job_id):
    """Get list of URLs to fetch."""

@hookspec(firstresult=True)
def before_fetch_url(conn, config, job_id, url, depth, request_headers):
    """Reject a URL, or modify its request headers."""

@hookspec(firstresult=True)
def fetch_cached_url(conn, config, job_id, url, depth, request_headers):
    """Fetch a previously cached URL."""

@hookspec()
def after_fetch_url(conn, config, job_id, url, request_headers, response, fresh, fetch_duration):
    """Process a fetched URL. Useful for caching or logging."""

@hookspec(firstresult=True)
def fetch_url(url, request_headers):
    """Fetch a URL live from an origin server."""

@hookspec()
def discover_urls(conn, config, job_id, url, response):
    """Discover new URLs to crawl."""

@hookspec()
def canonicalize_url(conn, config, job_id, from_url, to_url, to_url_depth):
    """Canonicalize a discovered URL, possibly rejecting it."""

@hookspec()
def extract_from_response(conn, config, job_id, url, response):
    """Extract some table rows from the response."""



from ..hookspecs import hookimpl
from datetime import datetime
import httpx

timeout = httpx.Timeout(5.0, read=20.0)
client = httpx.Client(timeout=timeout)

@hookimpl(trylast=True)
def fetch_url(url, request_headers):
    fetched_at = datetime.utcnow().isoformat(sep=' ')
    try:
        response = client.get(url, headers=request_headers)

        headers = []
        for k, v in response.headers.items():
            headers.append([k, v])
        return {
            'fetched_at': fetched_at,
            'headers': headers,
            'status_code': response.status_code,
            'text': response.text,
        }
    except Exception as e:
        return e



import time
import os
import json
import math
from urllib.parse import urljoin
from datasette_scraper.plugin import pm
from datasette_scraper import utils, inserts

def entrypoint_crawl(enabled_dbs, db_map):
    raw_factory = utils.lazy_connection_factory(db_map)
    conns = []
    for db in enabled_dbs:
        conns.append(raw_factory(db))

    try:
        while True:
            all_none = True
            for db in enabled_dbs:
                factory = utils.lazy_connection_factory_with_default(raw_factory, db)
                rv = crawl_loop(db, factory)

                if rv != None:
                    all_none = False

            if all_none:
                time.sleep(0.05)
    finally:
        for conn in conns:
            conn.close()

def get_next_crawl_queue_row(conn):
    # TODO: consider choosing randomly from the next N eligible items, this decreases
    #       contention so we have fewer failed attempts to claim a row in the
    #       entirely-cached path.
    #
    # I'm not clear on how beneficial this is in real life workloads.
    # e.g. on my 16-core machine, I can process 23K URLs in 24 seconds,
    # so the overhead of the system is ~16 CPU-milliseconds per URL. If your
    # code to process an item is of similar duration or longer, we won't get much
    # value from optimizing the infrastructure further.

    #row = conn.execute("WITH xs AS (SELECT id, job_id, host, queued_at, url, depth, claimed_at FROM dss_crawl_queue WHERE host IN (SELECT host FROM dss_host_rate_limit WHERE next_fetch_at < strftime('%Y-%m-%d %H:%M:%f')) AND (claimed_at IS NULL OR claimed_at < datetime('now', '-5 minutes')) ORDER BY depth, queued_at LIMIT 32) SELECT * FROM xs ORDER BY RANDOM() LIMIT 1").fetchone()
    row = conn.execute("SELECT id, job_id, host, queued_at, url, depth, claimed_at FROM dss_crawl_queue WHERE host IN (SELECT host FROM dss_host_rate_limit WHERE next_fetch_at < strftime('%Y-%m-%d %H:%M:%f')) AND (claimed_at IS NULL OR claimed_at < datetime('now', '-5 minutes')) ORDER BY depth, queued_at LIMIT 1").fetchone()

    if not row:
        # NB: None (as opposed to other falsy values) signals that there was no work,
        #     and we should potentially delay before querying again.
        return None

    id, job_id, host, queued_at, url, depth, claimed_at = row

    #print('...attempting to claim item id={}'.format(id))
    # Try to claim the row -- note that another worker may have claimed it while we were waiting.
    with conn:
        if claimed_at:
            claim_row = conn.execute("UPDATE dss_crawl_queue SET claimed_at = strftime('%Y-%m-%d %H:%M:%f') WHERE id = ? AND claimed_at = ?", [id, claimed_at])
        else:
            claim_row = conn.execute("UPDATE dss_crawl_queue SET claimed_at = strftime('%Y-%m-%d %H:%M:%f') WHERE id = ? AND claimed_at IS NULL", [id])

        if claim_row.rowcount != 1:
            #print('...failed to claim crawl item job_id={} id={} url={}'.format(job_id, id, url))
            return False

    return row

def absolutize_urls(base_url, new_url):
    try:
        rv = urljoin(base_url, new_url)
        hash_idx = rv.find('#')

        if hash_idx == -1:
            return rv

        return rv[0:hash_idx]
    except:
        return None

def update_crawl_stats(conn, job_id, host, response, fresh):
    with conn:
        conn.execute('BEGIN TRANSACTION')
        conn.execute("INSERT OR IGNORE INTO dss_job_stats(job_id, host) VALUES (?, ?)", [job_id, host])
        xx_column = 'fetched_5xx'
        status_code = response['status_code']

        if status_code >= 200 and status_code <= 299:
            xx_column = 'fetched_2xx'
        elif status_code >= 300 and status_code <= 399:
            xx_column = 'fetched_3xx'
        elif status_code >= 400 and status_code <= 499:
            xx_column = 'fetched_4xx'

        maybe_fetched_fresh = ''
        if fresh:
            maybe_fetched_fresh = ', fetched_fresh = fetched_fresh + 1'

        conn.execute("UPDATE dss_job_stats SET fetched = fetched + 1, {} = {} + 1{} WHERE job_id = ? AND host = ?".format(xx_column, xx_column, maybe_fetched_fresh), [job_id, host])

def discover_urls(conn, config, job_id, from_url, from_depth, response):
    urls = [new_url for urls in pm.hook.discover_urls(conn=conn, config=config, job_id=job_id, url=from_url, response=response) for new_url in urls]

    # Normalize URLs into (url, depth) form
    urls = [new_url if isinstance(new_url, tuple) else (new_url, from_depth + 1) for new_url in urls]

    # Resolve relative paths
    urls = [(absolutize_urls(from_url, new_url), new_depth) for (new_url, new_depth) in urls]
    urls = [x for x in urls if x[0]]

    # Reject non HTTP/HTTPS URLs
    urls = [new_url for new_url in urls if new_url[0].startswith('https:') or new_url[0].startswith('http:')]

    new_urls = []
    for (to_url, to_url_depth) in urls:
        attempts = 0
        while attempts < 10:
            # We do max 10 canonicalization attempts to prevent infinite loops from broken
            # plugins.
            attempts += 1
            results = pm.hook.canonicalize_url(conn=conn, config=config, job_id=job_id, from_url=from_url, to_url=to_url, to_url_depth=to_url_depth)

            rewritten = False
            for x in results:
                if isinstance(x, str) and x != to_url:
                    to_url = x
                    rewritten = True
                    break
                if isinstance(x, tuple) and x[0] != to_url and x[1] != to_url_depth:
                    to_url = x[0]
                    to_url_depth = x[1]
                    rewritten = True
                    break

            if rewritten:
                continue

            if False in results:
                # Someone rejected the URL; this wins.
                break

            new_urls.append((to_url, to_url_depth))
            break

    return set(list(new_urls))

def crawl_loop(dss_db_name, factory):
    """Try to process 1 pending URL. Return None to indicate there was no work to be done."""
    #print('crawl_loop running pid={}'.format(os.getpid()))
    # Warning: This is super naive! As we get experience operating at higher volumes,
    # may need to tweak it.

    # Try to find a URL to crawl that (a) isn't host rate limited and (b) hasn't been
    # recently claimed by another worker.
    conn = factory(None)

    row = get_next_crawl_queue_row(conn)

    if not row:
        return row

    id, job_id, host, queued_at, url, depth, claimed_at = row
    #print('! found work item {}'.format(row))
    config = utils.get_crawl_config_for_job_id(conn, job_id)

    # before_fetch_url: Give plugins a chance to reject this URL / add
    #  request headers.
    request_headers = {}
    rejected_reason = pm.hook.before_fetch_url(conn=conn, config=config, job_id=job_id, url=url, depth=depth, request_headers=request_headers)

    if rejected_reason:
        utils.reject_crawl_queue_item(conn, id, rejected_reason)
        utils.check_for_job_complete(conn, job_id)
        return True

    fresh = False
    start = time.time()
    # fetch_cached_url: Fetch a previously cached URL.
    response = pm.hook.fetch_cached_url(conn=conn, config=config, job_id=job_id, url=url, depth=depth, request_headers=request_headers)
    fetch_duration = math.ceil(1000 * (time.time() - start))

    if not response:
        # Try to update dss_host_rate_limit
        with conn:
            conn.execute('BEGIN TRANSACTION')
            claim_rate_limit = conn.execute("UPDATE dss_host_rate_limit SET next_fetch_at = strftime('%Y-%m-%d %H:%M:%f', 'now', delay_seconds || ' seconds') WHERE host = ? AND next_fetch_at < strftime('%Y-%m-%d %H:%M:%f')", [host])

            if claim_rate_limit.rowcount != 1:
                # ...we failed to get rate limit, so release the row
                print('...failed to claim host rate limit for job_id={} host={}'.format(job_id, host))
                # Release the row we were working on
                conn.execute('UPDATE dss_crawl_queue SET claimed_at = NULL WHERE id = ?', [id])
                return True


        fresh = True
        start = time.time()
        # fetch_url: Fetch the actual URL.
        response = pm.hook.fetch_url(url=url, request_headers=request_headers)

        if not response:
            # Weird, this should be impossible.
            utils.reject_crawl_queue_item(conn, id, 'fetch_url failed')
            utils.check_for_job_complete(conn, job_id)
            return True

        if isinstance(response, Exception):
            utils.reject_crawl_queue_item(conn, id, repr(response))
            utils.check_for_job_complete(conn, job_id)
            return True

        fetch_duration = math.ceil(1000 * (time.time() - start))

    # TODO: can we batch these updates? on a 20K crawl, I see a 15% savings if we don't
    #       call this. Since this is just statistics reporting, it'd be safe to defer
    #       writing, eg at most once every 5 seconds
    update_crawl_stats(conn, job_id, host, response, fresh)

    pm.hook.after_fetch_url(conn=conn, config=config, job_id=job_id, url=url, request_headers=request_headers, response=response, fresh=fresh, fetch_duration=fetch_duration)

    urls = discover_urls(conn, config, job_id, url, depth, response)
    # Try to insert these URLs into dss_crawl_queue; skipping entries that are already
    # present, or present in dss_crawl_queue_history with a lower or equal depth.
    now = time.time()
    enqueued = utils.add_crawl_queue_items(conn, job_id, urls)
    #print('enqueued {}/{} urls in {} sec'.format(enqueued, len(urls), time.time() - now))

    for insert in pm.hook.extract_from_response(conn=conn, config=config, job_id=job_id, url=url, response=response):
        if insert:
            rv = inserts.handle_insert(factory, insert)

            # TODO: this generates a lot of write txns and churn, but it's gloriously
            #       commutative. We should signal our counts to the coordinator,
            #       and it can update them on some cadence, eg once every second
            for ((dbname, tablename), (inserted, updated, deleted)) in rv.items():
                conn.execute('INSERT INTO dss_extract_stats(job_id, database, tbl, inserted, updated, deleted) VALUES (?, ?, ?, ?, ?, ?) ON CONFLICT(job_id, database, tbl) DO UPDATE SET inserted = inserted + ?, updated = updated + ?, deleted = deleted + ?', [job_id, dbname or dss_db_name, tablename, inserted, updated, deleted, inserted, updated, deleted])


    utils.finish_crawl_queue_item(conn, id, response, fresh, fetch_duration)
    utils.check_for_job_complete(conn, job_id)
    return True



from ..hookspecs import hookimpl
import re

DISCOVER_DENY = 'discover-deny'

_re = {}

@hookimpl
def canonicalize_url(config, from_url, to_url, to_url_depth):
    if not DISCOVER_DENY in config:
        return

    reject_res = config[DISCOVER_DENY]

    for reject_re in reject_res:
        if reject_re in _re:
            compiled = _re[reject_re]
        else:
            compiled = re.compile(reject_re)
            _re[reject_re] = compiled

        if compiled.search(to_url):
            return False

@hookimpl
def config_schema():
    from .. import ConfigSchema
    return ConfigSchema(
        schema = {
          'type': 'array',
          'items': {
              'type': 'string',
          }
        },
        uischema = {
            "type": "Control",
            "scope": '#/properties/{}'.format(DISCOVER_DENY),
            'label': "Never crawl URLs matching these regexes"
        },
        key = DISCOVER_DENY,
        sort = 3000,
        group = 'Links',
    )

@hookimpl
def config_default_value():
    return [
        # Try to avoid binary file types
        '(?i)[.](bmp|gif|gz|jpg|jpeg|m4a|m4v|mov|mpg|mpeg|mp3|mp4|pdf|png|tar|webp|wma|wmv|xls|xlsx|xz|zip)'
    ]



import sqlite3
from urllib.parse import urlparse
import time
import json
from datasette_scraper import ipc
from datasette_scraper.plugin import pm
from datasette_scraper.zstd import train_zstd_dict, get_active_dict_id, get_compressor, get_decompressor
from datasette_scraper.utils import lazy_connection_factory, get_crawl_config_for_job_id

def queue_discover_missing_dictionaries(conn):
    with conn:
        conn.execute('INSERT INTO dss_ops(type) VALUES (?)', [ipc.DISCOVER_MISSING_DICTIONARIES])

def entrypoint_ops(enabled_dbs, db_map):
    raw_factory = lazy_connection_factory(db_map)

    for db in enabled_dbs:
        conn = raw_factory(db)
        queue_discover_missing_dictionaries(conn)

    while True:
        for db in enabled_dbs:
            conn = raw_factory(db)
            next_op = conn.execute('SELECT id, type, config FROM dss_ops WHERE finished_at IS NULL ORDER BY id LIMIT 1').fetchone()

            if not next_op:
                time.sleep(0.1)
                continue

            op_id, type, config = next_op
            config = json.loads(config)
            print('dss_ops: db={} id={} type={} config={}'.format(db, op_id, type, config))

            if type == ipc.SEED_CRAWL:
                seed_crawl(conn, config['job-id'])
            elif type == ipc.DISCOVER_MISSING_DICTIONARIES:
                discover_missing_dictionaries(conn)
            elif type == ipc.TRAIN_DICTIONARY:
                train_dictionary(conn, config['host'])
            elif type == ipc.RECOMPRESS_ALL:
                recompress_all(conn)
            elif type == ipc.RECOMPRESS_HOST:
                recompress_host(conn, config['host'])
            elif type == ipc.RECOMPRESS_HASH:
                recompress_hash(conn, config['request_hash'])
            else:
                print('unknown dss_ops: id={} type={} config={}'.format(op_id, type, config))
            with conn:
                conn.execute("UPDATE dss_ops SET finished_at = strftime('%Y-%m-%d %H:%M:%f') WHERE id = ?", [op_id])

def recompress_all(conn):
    hosts = conn.execute('SELECT host FROM dss_fetch_cache GROUP BY 1').fetchall()

    with conn:
        for host in hosts:
            host = host[0]
            config = json.dumps({'host': host})
            conn.execute(
                'INSERT INTO dss_ops(type, config) SELECT ?, ?',
                [
                    ipc.RECOMPRESS_HOST,
                    config,
                ]
            )

def recompress_host(conn, host):
    hashes = conn.execute('SELECT request_hash FROM dss_fetch_cache WHERE host = ?', [host]).fetchall()

    with conn:
        for request_hash in hashes:
            request_hash = request_hash[0]
            config = json.dumps({'request_hash': request_hash})
            conn.execute(
                'INSERT INTO dss_ops(type, config) VALUES (?, ?)',
                [
                    ipc.RECOMPRESS_HASH,
                    config,
                ]
            )

def recompress_hash(conn, request_hash):
    row = conn.execute('SELECT url, object, dict_id FROM dss_fetch_cache WHERE request_hash = ?', [request_hash]).fetchone()

    if not row:
        return

    url, obj, old_dict_id = row

    host = urlparse(url).hostname

    dict_id = get_active_dict_id(conn, host)

    if dict_id == old_dict_id:
        return

    decompressed = get_decompressor(conn, old_dict_id).decompress(obj)

    compressed = get_compressor(conn, dict_id).compress(decompressed)

    with conn:
        conn.execute('UPDATE dss_fetch_cache SET object = ?, dict_id = ? WHERE request_hash = ?', [compressed, dict_id, request_hash])

def train_dictionary(conn, host):
    # Confirm that we still lack a dictionary
    exists, = conn.execute('SELECT EXISTS(SELECT * FROM dss_zstd_dict WHERE active AND host = ?)', [host]).fetchone()

    if exists:
         return

    train_zstd_dict(conn, host)
    with conn:
        conn.execute(
            'INSERT INTO dss_ops(type, config) VALUES (?, ?)',
            [ipc.RECOMPRESS_HOST, json.dumps({'host': host})]
        )

def discover_missing_dictionaries(conn):
    hosts = conn.execute('SELECT host FROM dss_fetch_cache GROUP BY 1 HAVING COUNT(*) > 100 EXCEPT SELECT host FROM dss_zstd_dict').fetchall()

    for host in hosts:
        host = host[0]
        with conn:
            config = json.dumps({'host': host})
            conn.execute(
                'INSERT INTO dss_ops(type, config) SELECT ?, ? WHERE NOT EXISTS(SELECT * FROM dss_ops WHERE finished_at IS NULL AND type = ? AND config = ?)',
                [
                    ipc.TRAIN_DICTIONARY,
                    config,
                    ipc.TRAIN_DICTIONARY,
                    config
                ]
            )

def seed_crawl(conn, job_id):
    config = get_crawl_config_for_job_id(conn, job_id)
    seeds = [url for urls in pm.hook.get_seed_urls(conn=conn, config=config, job_id=job_id) for url in urls]

    hosts = []
    with conn:
        for seed in seeds:
            url = urlparse(seed)
            hostname = url.hostname
            hosts.append(hostname)
            conn.execute('INSERT INTO dss_crawl_queue(job_id, host, url, depth) VALUES (?, ?, ?, 0)', [job_id, hostname, seed])

        for host in set(hosts):
            conn.execute('INSERT INTO dss_host_rate_limit(host) SELECT ? WHERE NOT EXISTS(SELECT * FROM dss_host_rate_limit WHERE host = ?)', [host, host])

        conn.execute("UPDATE dss_job SET status = 'running' WHERE id = ?", [job_id])



import sqlite3
from .migrator import DBMigrator
from .schema import current_schema_version, schema
from .errors import ScraperError

_plugin_name = 'datasette-scraper'

_enabled_databases = None

def enabled_databases(datasette, empty_if_not_initialized=False):
    global _enabled_databases

    if not _enabled_databases is None:
        return _enabled_databases

    if empty_if_not_initialized:
        return []

    global_config = datasette.plugin_config(_plugin_name)

    rv = []

    for db_name in datasette.databases:
        local_config = datasette.plugin_config(_plugin_name, db_name)

        if local_config is None:
            continue

        rv.append(db_name)

    _enabled_databases = rv
    return _enabled_databases

async def get_db_version(db):
    results = await db.execute('pragma user_version')
    for row in results:
        return row['user_version']

def ensure_wal_mode(conn):
    old_level = conn.isolation_level
    try:
        conn.isolation_level = None
        mode, = conn.execute('PRAGMA journal_mode=WAL').fetchone()
        if mode != 'wal':
            raise Exception('unable to set PRAGMA journal_mode=WAL on connection, got {}'.format(mode))
    finally:
        conn.isolation_level = old_level

async def ensure_schema(db):
    def ensure_schema_internal(conn):
        ensure_wal_mode(conn)

        v, = conn.execute("PRAGMA user_version").fetchone()

        if not v:
            print('Installing datasette-scraper schema into db {}'.format(db.name))
            #conn.execute('create table foo(version int)')
            with DBMigrator(conn, schema, allow_deletions=True) as migrator:
                migrator.migrate()
        elif v == current_schema_version:
            pass
        elif v == 1000000 or v == 1000001 or v == 1000002:
            # Nothing special required - these just added tables/columns/indexes
            with DBMigrator(conn, schema, allow_deletions=True) as migrator:
                migrator.migrate()
        else:
            raise ScraperError('unsupported schema version in db {}: {} -- you may need to give datasette-scraper its own database'.format(db.name, v))


    await db.execute_write_fn(ensure_schema_internal, block=True)
    version = await get_db_version(db)

    if version != current_schema_version:
        raise ScraperError('unable to ensure schema in database {} (version={}; desired={}); please check that the database is mutable and not the _memory database'.format(db.name, version, current_schema_version))

    names = await db.table_names()
    # TODO: We should loop over our configs and enable WAL mode in any target database.
    #       Maybe there's a way to do that lazily, once?




from ..hookspecs import hookimpl
from ..utils import get_html_parser
import re

DISCOVER_HTML_LINKS = 'discover-html-links'

_re = {}

def discover(response, url, options):
    # Only scan for links if url-regex is absent, or matches url
    if 'url-regex' in options:
        regex = options['url-regex']
        if regex in _re:
            compiled = _re[regex]
        else:
            compiled = re.compile(regex)
            _re[regex] = compiled

        if not compiled.search(url):
            return []

    parsed = get_html_parser(response)
    els = parsed.css(options['selector'])
    attribute = None
    if 'attribute' in options:
        attribute = options['attribute']

    depth = None
    if 'depth' in options:
        depth = options['depth']

    rv = []
    for p in els:
        needle = ''
        if attribute:
            attrs = p.attributes
            if attribute in attrs:
                needle = attrs[attribute]
        else:
            needle = p.text()


        if needle and needle.strip():
            if depth == None:
                rv.append(needle.strip())
            else:
                rv.append((needle.strip(), depth))

    return rv

@hookimpl
def discover_urls(config, url, response):
    if not DISCOVER_HTML_LINKS in config:
        return []

    sources = config[DISCOVER_HTML_LINKS]

    rv = []

    for source in sources:
        rv.extend(discover(response, url, source))

    return list(set(rv))

@hookimpl
def config_schema():
    from .. import ConfigSchema
    return ConfigSchema(
        schema = {
          'type': 'array',
          'items': {
              'type': 'object',
              'properties': {
                  'url-regex': {
                      'type': 'string',
                      'title': 'On pages with URLs matching...',
                  },
                  'selector': {
                      'type': 'string',
                      'default': 'a',
                      'title': 'CSS Selector',
                      'description': 'description',
                      'examples': ['example']
                  },
                  'attribute': {
                      'type': 'string',
                      'title': 'HTML Attribute',
                  },
                  'depth': {
                      'type': 'integer',
                      'title': 'Depth',
                  }
              },
              'required': ['selector']
          }
        },
        uischema = {
            "type": "Control",
            "scope": '#/properties/{}'.format(DISCOVER_HTML_LINKS),
            'label': 'Find links in HTML content',
        },
        key = DISCOVER_HTML_LINKS,
        sort = 2000,
        group = 'Links',
    )

@hookimpl
def config_default_value():
    return [{
        'selector': 'a',
        'attribute': 'href',
    }]



from ..hookspecs import hookimpl

MAX_DEPTH = 'max-depth'

@hookimpl
def canonicalize_url(config, to_url_depth):
    if MAX_DEPTH in config:
        max_depth = config[MAX_DEPTH]

        if to_url_depth > max_depth:
            return False

@hookimpl
def config_schema():
    from .. import ConfigSchema
    return ConfigSchema(
        schema = {
          'type': 'number',
        },
        uischema = {
            "type": "Control",
            "scope": '#/properties/{}'.format(MAX_DEPTH)
        },
        key = MAX_DEPTH,
        group = 'Links',
        sort = 100,
    )

@hookimpl
def config_default_value():
    return 3



from datasette.app import Datasette
import sqlite3
import pytest

import datasette_scraper.config


@pytest.mark.asyncio
async def test_plugin_is_installed(tmp_path):
    # Ensure we don't hit the cached values
    datasette_scraper.config._enabled_databases = None
    db_name = tmp_path / "db.sqlite"
    conn = sqlite3.connect(db_name)
    conn.close()
    datasette = Datasette(files=[db_name])
    response = await datasette.client.get("/-/plugins.json")
    assert response.status_code == 200
    installed_plugins = {p["name"] for p in response.json()}
    assert "datasette-scraper" in installed_plugins

@pytest.mark.asyncio
async def test_plugin_tolerates_existing_tables(tmp_path):
    # Ensure we don't hit the cached values
    datasette_scraper.config._enabled_databases = None

    metadata = {
        'databases': {
            'db': {
                'plugins': {
                    'datasette-scraper': {
                    }
                }
            }
        }
    }

    db_name = tmp_path / "db.sqlite"
    conn = sqlite3.connect(db_name)
    with conn:
        conn.execute('CREATE TABLE x(x TEXT NOT NULL PRIMARY KEY)')
        conn.execute('CREATE UNIQUE INDEX y ON x(x)')
    conn.close()
    datasette = Datasette(
        files=[db_name],
        metadata=metadata
    )
    response = await datasette.client.get("/-/plugins.json")
    assert response.status_code == 200
    installed_plugins = {p["name"] for p in response.json()}
    assert "datasette-scraper" in installed_plugins



current_schema_version = 1000003;

schema = """
PRAGMA user_version = {};
""".format(current_schema_version) + """

-- Site-specific dictionaries for zstd
CREATE TABLE dss_zstd_dict(
  id integer primary key,
  host text not null,
  active boolean not null default true,
  dict blob not null
);


-- Global rate limits. Currently, no attempt at fairness - a run of
-- a crawl can starve runs of other crawls.
CREATE TABLE dss_host_rate_limit(
  host text primary key,
  next_fetch_at text not null default (strftime('%Y-%m-%d %H:%M:%f')),
  delay_seconds integer not null default 10
);

-- A cache of fetched pages. This may need to exist outside of
-- the schema migration code if we adopt https://github.com/phiresky/sqlite-zstd
CREATE TABLE dss_fetch_cache(
  -- A hash of the request that was sent, eg url + headers
  request_hash text primary key,

  host text not null,

  -- The URL that was fetched
  url text not null,

  -- When this was fetched from an origin server. This will also be
  -- present in the object, but exposing it as a column makes it
  -- easier to prune old entries.
  fetched_at text not null default (strftime('%Y-%m-%d %H:%M:%f')),

  -- When this was last accessed via a crawl -- can be helpful for
  -- pruning entries that are no longer needed.
  read_at text not null default (strftime('%Y-%m-%d %H:%M:%f')),

  -- The response object; UTF-8 JSON encoded. Might be compressed.
  object blob not null,

  dict_id integer references dss_zstd_dict(id)
);

-- The definition of a crawl, eg "Crawl every page at most 3 hops away from
-- news.ycombinator.com"
CREATE TABLE dss_crawl(
  id integer primary key,
  name text not null,
  config text not null
);

-- A specific attempt to perform a crawl
CREATE TABLE dss_job(
  id integer primary key,
  crawl_id integer not null references dss_crawl(id) on delete cascade,
  started_at text not null default (strftime('%Y-%m-%d %H:%M:%f')),
  status text not null default 'seeding' check (status IN ('seeding', 'running', 'cancelled', 'done')),
  -- NULL finished-at => currently in progress
  finished_at text
);

-- URLs to be crawled
CREATE TABLE dss_crawl_queue(
  id integer primary key,
  job_id integer not null references dss_job(id) on delete cascade,
  host text not null,
  queued_at text not null default (strftime('%Y-%m-%d %H:%M:%f')),
  url text not null,
  depth integer not null,
  claimed_at text
);

-- URLs that were crawled
CREATE TABLE dss_crawl_queue_history(
  job_id integer not null references dss_job(id) on delete cascade,
  host text not null,
  url text not null,
  depth integer not null,
  processed_at text not null,
  fetched_fresh boolean not null,
  status_code integer,
  content_type text,
  size integer,
  duration integer,
  skipped_reason text,
  request_hash text references dss_fetch_cache(request_hash) on delete cascade,
  check (
    (skipped_reason is null and status_code is not null and size is not null and content_type is not null and duration is not null) or
    (skipped_reason is not null and status_code is null and size is null and content_type is null and duration is null)
  )
);

-- Metrics on a specific crawl
CREATE TABLE dss_job_stats(
  job_id integer not null references dss_job(id) on delete cascade,
  host text not null,
  fetched integer not null default 0,
  fetched_fresh integer not null default 0,
  fetched_2xx integer not null default 0,
  fetched_3xx integer not null default 0,
  fetched_4xx integer not null default 0,
  fetched_5xx integer not null default 0,
  primary key (job_id, host)
);

CREATE TABLE dss_extract_stats(
  job_id integer not null references dss_job(id) on delete cascade,
  database text not null,
  tbl text not null,
  inserted integer not null default 0,
  updated integer not null default 0,
  deleted integer not null default 0,
  primary key (job_id, database, tbl)
);

CREATE TABLE dss_ops(
  id integer primary key,
  queued_at text not null default (strftime('%Y-%m-%d %H:%M:%f')),
  finished_at text,
  type text not null,
  config text not null default '{}'
);

CREATE UNIQUE INDEX idx_only_one_active_job_per_crawl ON dss_job(crawl_id) WHERE finished_at IS NULL;

CREATE INDEX idx_crawl_queue_item ON dss_crawl_queue(job_id, url);
CREATE INDEX idx_crawl_queue_history_item ON dss_crawl_queue_history(job_id, url);
CREATE INDEX idx_dss_zstd_dict_host ON dss_zstd_dict(host);
CREATE INDEX idx_dss_fetch_cache_host ON dss_fetch_cache(host);
CREATE INDEX idx_crawl_queue_host_depth_queued_at ON dss_crawl_queue(host, depth, queued_at);
"""



from ..hookspecs import hookimpl
from urllib.parse import urlparse

DISCOVER_ONLY_SAME_ORIGIN = 'discover-only-same-origin'

@hookimpl
def canonicalize_url(config, from_url, to_url, to_url_depth):
    if not DISCOVER_ONLY_SAME_ORIGIN in config or not config[DISCOVER_ONLY_SAME_ORIGIN]:
        return

    from_ = urlparse(from_url)
    to_ = urlparse(to_url)

    return from_.hostname == to_.hostname

@hookimpl
def config_schema():
    from .. import ConfigSchema
    return ConfigSchema(
        schema = {
          'type': 'boolean',
          'title': 'Stay on same domain',
        },
        uischema = {
            "type": "Control",
            "scope": '#/properties/{}'.format(DISCOVER_ONLY_SAME_ORIGIN)
        },
        key = DISCOVER_ONLY_SAME_ORIGIN,
        group = 'Links',
        sort = 0
    )

@hookimpl
def config_default_value():
    return True



from ..hookspecs import hookimpl

SEED_URLS = 'seed-urls'

@hookimpl
def get_seed_urls(config):
    if SEED_URLS in config:
        return config[SEED_URLS]

    return []

@hookimpl
def config_schema():
    from .. import ConfigSchema
    return ConfigSchema(
        schema = {
          'type': 'array',
          'items': {
            'type': 'string',
              'pattern': '^https?://.+'
          }
        },
        uischema = {
            "type": "Control",
            "scope": '#/properties/{}'.format(SEED_URLS),
            'label': 'URLs'
        },
        key = SEED_URLS,
        group = 'Seeds',
    )

@hookimpl
def config_default_value():
    return []



import re
from ..hookspecs import hookimpl
from .seed_urls import SEED_URLS
from ..utils import get_html_parser
from urllib.parse import urlparse

SEED_SITEMAPS = 'seed-sitemaps'
cdata_re = re.compile('<!--\\[CDATA\\[(.+)]]-->')

@hookimpl
def get_seed_urls(config):
    if not SEED_SITEMAPS in config or not config[SEED_SITEMAPS]:
        return []

    rv = []
    seen_hosts = {}
    # We cheat, and peek into the config for seed-urls to find domains whose
    # robots.txt we'll sniff for a sitemap.
    seed_urls = []
    if SEED_URLS in config:
        seed_urls = config[SEED_URLS]

    for url in seed_urls:
        parsed = urlparse(url)
        host = parsed.hostname

        if host in seen_hosts:
            continue

        # Only crawl a sitemap if the seed is for a root.
        # This lets users mix-and-match directed crawls
        # with sitemap crawls.
        if parsed.path != '' and parsed.path != '/':
            continue

        seen_hosts[host] = True

        candidates = [
            '{}://{}/robots.txt'.format(parsed.scheme, host),
            '{}://{}/sitemap.xml'.format(parsed.scheme, host),
            '{}://{}/sitemap_index.xml'.format(parsed.scheme, host)
        ]

        for candidate in candidates:
            if not candidate in seed_urls:
                rv.append(candidate)

    return rv



def extract_text(node):
    # selectolax wraps CDATA declarations in comments
    url = node.text()
    if url:
        return url

    child = node.child

    if child and child.tag == '_comment':
        m = cdata_re.search(child.html)

        if m:
            return m.group(1)


@hookimpl
def discover_urls(config, url, response):
    if not SEED_SITEMAPS in config or not config[SEED_SITEMAPS]:
        return []

    parsed = urlparse(url)

    if parsed.path == '/robots.txt':
        sitemaps = list(set([x.split(' ')[1] for x in response['text'].split('\n') if x.startswith('Sitemap: ')]))
        return sitemaps

    is_https = url.startswith('https://')

    # Some plugins generate HTTP sitemap URLs even when served on HTTPS.
    # Detect and fix that.
    def fixup(old):
        if is_https and old.startswith('http://'):
            return old.replace('http://', 'https://')

        return old

    rv = []
    if parsed.path.endswith('.xml'):
        parsed = get_html_parser(response)

        for node in parsed.css('sitemapindex sitemap loc'):
            url = extract_text(node)
            if url:
                rv.append((fixup(url), 0))

        for node in parsed.css('urlset url loc'):
            url = extract_text(node)
            if url:
                rv.append((fixup(url), 1))

    return rv

@hookimpl
def config_schema():
    from .. import ConfigSchema
    return ConfigSchema(
        schema = {
          'type': 'boolean',
            'title': 'Follow XML sitemaps',
        },
        uischema = {
            "type": "Control",
            "scope": '#/properties/{}'.format(SEED_SITEMAPS),
        },
        key = SEED_SITEMAPS,
        group = 'Seeds',
        sort = 100,
    )

@hookimpl
def config_default_value():
    return True



import datasette
import html
import glob
from jinja2 import FunctionLoader
import re
import markupsafe
import os
import json
from .hookspecs import hookimpl
from .config import enabled_databases, ensure_schema
from .plugin import pm
from .routes import get_routes
from .workers import start_workers
from .utils import module_from_path
from collections import namedtuple

ConfigSchema = namedtuple('ConfigSchema', ['schema', 'uischema', 'key', 'group', 'sort'], defaults=(None, 1000))

class JsonString:
    def __init__(self, obj):
        self.obj = obj

    def __html__(self):
        return json.dumps(self.obj)


@datasette.hookimpl
def startup(datasette):
    async def inner():
        enabled_databases = config.enabled_databases(datasette)
        for db_name in enabled_databases:
            await ensure_schema(datasette.databases[db_name])

        if enabled_databases:
            # TODO: this isn't a documented surface area, so it's a bit skeezy to be using it
            if datasette.plugins_dir:
                for filepath in glob.glob(os.path.join(datasette.plugins_dir, "*.py")):
                    if not os.path.isfile(filepath):
                        continue
                    mod = module_from_path(filepath, name=os.path.basename(filepath))
                    try:
                        pm.register(mod)
                    except ValueError:
                        # Plugin already registered
                        pass

            start_workers(datasette)


    return inner

class MyFunctionLoader(FunctionLoader):
    def list_templates(self):
        return []

@datasette.hookimpl
def prepare_jinja2_environment(env, datasette):
    def load_func(path):
        try:
            if path.startswith('table-') and path.endswith('-dss_crawl.html'):
                path = 'table-dss_crawl.html'
            elif path.startswith('table-') and path.endswith('-dss_job.html'):
                path = 'table-dss_job.html'
            elif path.startswith('table-') and path.endswith('-dss_host_rate_limit.html'):
                path = 'table-dss_host_rate_limit.html'
            elif path.startswith('table-') and path.endswith('-dss_job_stats.html'):
                path = 'table-dss_job_stats.html'
            elif path.startswith('table-') and path.endswith('-dss_extract_stats.html'):
                path = 'table-dss_extract_stats.html'
            elif path.startswith('table-') and path.endswith('-dss_zstd_dict.html'):
                path = 'table-dss_zstd_dict.html'
            elif path.startswith('row-') and path.endswith('-dss_crawl.html'):
                path = 'row-dss_crawl.html'
            elif path.startswith('row-') and path.endswith('-dss_host_rate_limit.html'):
                path = 'row-dss_host_rate_limit.html'

            template_path = os.path.abspath(os.path.join(__file__, '..', 'templates', path))

            f = open(template_path, 'r')
            contents = f.read()
            f.close()

            # TODO: this should return True in prod, I think? Although maybe
            #       just a performance improvement, so it's fine.
            return contents, path, lambda: False
        except FileNotFoundError:
            return None

    env.loader.loaders.insert(0, MyFunctionLoader(load_func))

@datasette.hookimpl
def get_metadata(datasette, key, database, table):
    rv = {
        'databases': {}
    }

    # There's a weird circular dependency issue here.
    # When we called enabled_databases, it'll check plugin
    # configuration to see if it's enabled.
    #
    # Which means it's not safe for _us_ to call enabled_databases
    # until _after_ it has initialized.
    for db_name in enabled_databases(datasette, empty_if_not_initialized=True):
        rv['databases'][db_name] = {
            'tables': {
                'dss_crawl_queue': {
                    'sort_desc': 'id'
                },
                'dss_crawl_queue_history': {
                    'sort_desc': 'processed_at'
                },
                'dss_fetch_cache': {
                    'sort_desc': 'fetched_at'
                },
                'dss_job': {
                    'sort_desc': 'id'
                },
                'dss_job_stats': {
                    'sort_desc': 'job_id'
                },

            }
        }

    return rv

@datasette.hookimpl
def extra_template_vars(datasette, request):
    """Add dss_schema, dss_default_config, dss_id variables."""

    # TODO: this is pretty janky! It would be nice to add these only
    #       on our templates.

    known_groups = {
        'Seeds': 1,
        'Links': 2,
        'Caching': 5,
        'Extracting': 6,
        'Other': 7,
    }

    groups = {}

    async def extra_vars():
        schema = {
            'type': 'object',
            'properties': {
            }
        };

        schema['properties']['name'] = {
            'type': 'string',
            'minLength': 1
        }

        default_config = {'name': ''}

        for plugin in pm.get_plugins():
            if not 'config_schema' in dir(plugin):
                continue

            rv = plugin.config_schema()

            schema['properties'][rv.key] = rv.schema

            group = rv.group
            if not group in known_groups:
                group = 'Other'

            groups[group] = groups.get(group, [])
            groups[group].append((rv.uischema, rv.sort))

            if 'config_default_value' in dir(plugin):
                default_config[rv.key] = plugin.config_default_value()

        id = 0

        default_config['name'] = 'Crawl name'

        category_schemas = []
        for key in sorted(groups.keys(), key=lambda x: known_groups[x]):
            elements = sorted(groups[key], key=lambda x: x[1])
            elements = [x[0] for x in elements]
            category_schemas.append({
                'type': 'Category',
                'label': key,
                'elements': elements
            })

        uischema = {
            "type": "VerticalLayout",
            "elements": [
                {
                    "type": "Control",
                    "scope": "#/properties/name",
                },
                {
                    "type": "Categorization",
                    "elements": category_schemas,
                }
            ]
        }

        dss_db = None
        m = re.search('^/([^/]+)/', request.path)
        if m:
            dss_db = m.group(1)

        m = re.search('/([^/]+)/dss_crawl/([0-9]+)$', request.path)
        if m:
            db_name = m.group(1)
            db = datasette.databases[db_name]
            id = int(m.group(2))
            rv = await db.execute('SELECT name, config FROM dss_crawl WHERE id = ?', [id])
            for row in rv:
                config = json.loads(row['config'])
                config['name'] = row['name']
                default_config = config

        return {
            "dss_schema": JsonString(schema),
            "dss_uischema": JsonString(uischema),
            "dss_default_config": JsonString(default_config),
            "dss_db": dss_db,
            "dss_id": id
        }

    return extra_vars

@datasette.hookimpl
def render_cell(database, row, table, column, value):
    if table != 'dss_job_stats' and table != 'dss_extract_stats':
        return None

    def link(label, href):
        return markupsafe.Markup(
            '<a href="{href}">{label}</a>'.format(
                href=markupsafe.escape(href),
                label=markupsafe.escape(label)
            )
        )

    if table == 'dss_job_stats':
        if column == 'host':
            return link(value, '/{}/dss_crawl_queue_history?job_id__exact={}&host__exact={}&_sort_desc=processed_at'.format(database, row['job_id']['value'], row['host']))
        elif column == 'fetched_fresh':
            return link(value, '/{}/dss_crawl_queue_history?job_id__exact={}&host__exact={}&fetched_fresh__exact=1&_sort_desc=processed_at'.format(database, row['job_id']['value'], row['host']))
        elif column.startswith('fetched_') and column.endswith('xx'):
            start = 100 * int(column[8])
            return link(value, '/{}/dss_crawl_queue_history?job_id__exact={}&host__exact={}&status_code__gte={}&status_code__lte={}&_sort_desc=processed_at'.format(database, row['job_id']['value'], row['host'], start, start + 99))

    if table == 'dss_extract_stats':
        if column == 'database':
            return link(value, '/{}'.format(row['database']))

        if column == 'tbl':
            return link(value, '/{}/{}'.format(row['database'], row['tbl']))


@datasette.hookimpl
def register_routes(datasette):
    return get_routes(datasette)



import importlib
import pluggy
import sys
from . import hookspecs

DEFAULT_PLUGINS = (
    "datasette_scraper.plugins.fetch_url",
    "datasette_scraper.plugins.fetch_cache",
    "datasette_scraper.plugins.discover_allow",
    "datasette_scraper.plugins.discover_deny",
    "datasette_scraper.plugins.discover_html_links",
    "datasette_scraper.plugins.discover_only_same_origin",
    "datasette_scraper.plugins.discover_maintain_ssl",
    "datasette_scraper.plugins.canonicalize_shopify_products",
    "datasette_scraper.plugins.follow_redirects",
    "datasette_scraper.plugins.seed_urls",
    "datasette_scraper.plugins.seed_sitemaps",
    "datasette_scraper.plugins.max_depth",
    "datasette_scraper.plugins.max_pages",
    "datasette_scraper.plugins.max_pages_per_domain",
    "datasette_scraper.plugins.extract_links",
    "datasette_scraper.plugins.extract_json_ld",
    "datasette_scraper.plugins.extract_json_ld_product",
)

pm = pluggy.PluginManager("datasette_scraper")
pm.add_hookspecs(hookspecs)

if not hasattr(sys, "_called_from_test"):
    # Only load plugins if not running tests
    pm.load_setuptools_entrypoints("datasette_scraper")

# Load default plugins
for plugin in DEFAULT_PLUGINS:
    mod = importlib.import_module(plugin)
    pm.register(mod, plugin)



# coding: utf-8

"""Simple declarative schema migration for SQLite.

See <https://david.rothlis.net/declarative-schema-migration-for-sqlite>.

Author: William Manley <will@stb-tester.com>.
Copyright © 2019-2022 Stb-tester.com Ltd.
License: MIT.
"""

# 2022-12-26: cldellow: This is patched from the original. It only
#                       considers db objects that begin with dss_.

import logging
import re
import sqlite3
from textwrap import dedent


def dumb_migrate_db(db, schema, allow_deletions=False):
    """
    Migrates a database to the new schema given by the SQL text `schema`
    preserving the data.  We create any table that exists in schema, delete any
    old table that is no longer used and add/remove columns and indices as
    necessary.

    Under this scheme there are a set of changes that we can make to the schema
    and this script will handle it fine:

    1. Adding a new table
    2. Adding, deleting or modifying an index
    3. Adding a column to an existing table as long as the new column can be
       NULL or has a DEFAULT value specified.
    4. Changing a column to remove NULL or DEFAULT as long as all values in the
       database are not NULL
    5. Changing the type of a column
    6. Changing the user_version

    In addition this function is capable of:

    1. Deleting tables
    2. Deleting columns from tables

    But only if allow_deletions=True.  If the new schema requires a column/table
    to be deleted and allow_deletions=False this function will raise
    `RuntimeError`.

    Note: When this function is called a transaction must not be held open on
    db.  A transaction will be used internally.  If you wish to perform
    additional migration steps as part of a migration use DBMigrator directly.

    Any internally generated rowid columns by SQLite may change values by this
    migration.
    """
    with DBMigrator(db, schema, allow_deletions) as migrator:
        migrator.migrate()
    return bool(migrator.n_changes)


class DBMigrator:
    def __init__(self, db, schema, allow_deletions=False):
        self.db = db
        self.schema = schema
        self.allow_deletions = allow_deletions

        self.pristine = sqlite3.connect(":memory:")
        self.pristine.executescript(schema)
        self.n_changes = 0

        self.orig_foreign_keys = None

    def log_execute(self, msg, sql, args=None):
        # It's important to log any changes we're making to the database for
        # forensics later
        msg_tmpl = "Database migration: %s with SQL:\n%s"
        msg_argv = (msg, _left_pad(dedent(sql)))
        if args:
            msg_tmpl += " args = %r"
            msg_argv += (args,)
        else:
            args = []
        logging.info(msg_tmpl, *msg_argv)
        self.db.execute(sql, args)
        self.n_changes += 1

    def __enter__(self):
        self.orig_foreign_keys = (
            self.db.execute("PRAGMA foreign_keys").fetchone()[0])
        if self.orig_foreign_keys:
            self.log_execute("Disable foreign keys temporarily for migration",
                             "PRAGMA foreign_keys = OFF")
            # This doesn't count as a change because we'll undo it at the end
            self.n_changes = 0

        self.db.__enter__()
        self.db.execute('BEGIN')
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.db.__exit__(exc_type, exc_value, exc_tb)
        if exc_value is None:
            # The SQLite docs say:
            #
            # > This pragma is a no-op within a transaction; foreign key
            # > constraint enforcement may only be enabled or disabled when
            # > there is no pending BEGIN or SAVEPOINT.
            old_changes = self.n_changes
            new_val = self._migrate_pragma('foreign_keys')
            if new_val == self.orig_foreign_keys:
                self.n_changes = old_changes

            # SQLite docs say:
            #
            # > A VACUUM will fail if there is an open transaction on the database
            # > connection that is attempting to run the VACUUM.
            if self.n_changes:
                self.db.execute("VACUUM")
        else:
            if self.orig_foreign_keys:
                self.log_execute(
                    "Re-enable foreign keys after migration",
                    "PRAGMA foreign_keys = ON")

    def migrate(self):
        # In CI the database schema may be changing all the time.  This checks
        # the current db and if it doesn't match database.sql we will
        # modify it so it does match where possible.
        pristine_tables = dict(self.pristine.execute("""\
            SELECT name, sql FROM sqlite_master
            WHERE type = \"table\" AND name != \"sqlite_sequence\" AND tbl_name LIKE \"dss_%\"""").fetchall())
        pristine_indices = dict(self.pristine.execute("""\
            SELECT name, sql FROM sqlite_master
            WHERE type = \"index\" AND tbl_name LIKE \"dss_%\"""").fetchall())

        tables = dict(self.db.execute("""\
            SELECT name, sql FROM sqlite_master
            WHERE type = \"table\" AND name != \"sqlite_sequence\" AND tbl_name LIKE \"dss_%\"""").fetchall())

        new_tables = set(pristine_tables.keys()) - set(tables.keys())
        removed_tables = set(tables.keys()) - set(pristine_tables.keys())
        if removed_tables and not self.allow_deletions:
            raise RuntimeError(
                "Database migration: Refusing to delete tables %r" %
                removed_tables)

        modified_tables = set(
            name for name, sql in pristine_tables.items()
            if normalise_sql(tables.get(name, "")) != normalise_sql(sql))

        # This PRAGMA is automatically disabled when the db is committed
        self.db.execute("PRAGMA defer_foreign_keys = TRUE")

        # New and removed tables are easy:
        for tbl_name in new_tables:
            self.log_execute("Create table %s" % tbl_name,
                             pristine_tables[tbl_name])
        for tbl_name in removed_tables:
            self.log_execute("Drop table %s" % tbl_name,
                             "DROP TABLE %s" % tbl_name)

        for tbl_name in modified_tables:
            # The SQLite documentation insists that we create the new table and
            # rename it over the old rather than moving the old out of the way
            # and then creating the new
            create_table_sql = pristine_tables[tbl_name]
            create_table_sql = re.sub(r"\b%s\b" % re.escape(tbl_name),
                                      tbl_name + "_migration_new",
                                      create_table_sql)
            self.log_execute(
                "Columns change: Create table %s with updated schema" %
                tbl_name, create_table_sql)

            cols = set([
                x[1] for x in self.db.execute(
                    "PRAGMA table_info(%s)" % tbl_name)])
            pristine_cols = set([
                x[1] for x in
                self.pristine.execute("PRAGMA table_info(%s)" % tbl_name)])

            removed_columns = cols - pristine_cols
            if not self.allow_deletions and removed_columns:
                logging.warning(
                    "Database migration: Refusing to remove columns %r from "
                    "table %s.  Current cols are %r attempting migration to %r",
                    removed_columns, tbl_name, cols, pristine_cols)
                raise RuntimeError(
                    "Database migration: Refusing to remove columns %r from "
                    "table %s" % (removed_columns, tbl_name))

            logging.info("cols: %s, pristine_cols: %s", cols, pristine_cols)
            self.log_execute(
                "Migrate data for table %s" % tbl_name, """\
                INSERT INTO {tbl_name}_migration_new ({common})
                SELECT {common} FROM {tbl_name}""".format(
                    tbl_name=tbl_name,
                    common=", ".join(cols.intersection(pristine_cols))))

            # Don't need the old table any more
            self.log_execute(
                "Drop old table %s now data has been migrated" % tbl_name,
                "DROP TABLE %s" % tbl_name)

            self.log_execute(
                "Columns change: Move new table %s over old" % tbl_name,
                "ALTER TABLE %s_migration_new RENAME TO %s" % (
                    tbl_name, tbl_name))

        # Migrate the indices
        indices = dict(self.db.execute("""\
            SELECT name, sql FROM sqlite_master
            WHERE type = \"index\" AND tbl_name LIKE 'dss_%'""").fetchall())
        for name in set(indices.keys()) - set(pristine_indices.keys()):
            self.log_execute("Dropping obsolete index %s" % name,
                             "DROP INDEX %s" % name)
        for name, sql in pristine_indices.items():
            if name not in indices:
                self.log_execute("Creating new index %s" % name, sql)
            elif sql != indices[name]:
                self.log_execute(
                    "Index %s changed: Dropping old version" % name,
                    "DROP INDEX %s" % name)
                self.log_execute(
                    "Index %s changed: Creating updated version in its place" %
                    name, sql)

        self._migrate_pragma('user_version')

        if self.pristine.execute("PRAGMA foreign_keys").fetchone()[0]:
            if self.db.execute("PRAGMA foreign_key_check").fetchall():
                raise RuntimeError(
                    "Database migration: Would fail foreign_key_check")

    def _migrate_pragma(self, pragma):
        pristine_val = self.pristine.execute(
            "PRAGMA %s" % pragma).fetchone()[0]
        val = self.db.execute("PRAGMA %s" % pragma).fetchone()[0]

        if val != pristine_val:
            self.log_execute(
                "Set %s to %i from %i" % (pragma, pristine_val, val),
                "PRAGMA %s = %i" % (pragma, pristine_val))

        return pristine_val


def _left_pad(text, indent="    "):
    """Maybe I can find a package in pypi for this?"""
    return "\n".join(indent + line for line in text.split('\n'))


def normalise_sql(sql):
    # Remove comments:
    sql = re.sub(r'--[^\n]*\n', "", sql)
    # Normalise whitespace:
    sql = re.sub(r'\s+', " ", sql)
    sql = re.sub(r" *([(),]) *", r"\1", sql)
    # Remove unnecessary quotes
    sql = re.sub(r'"(\w+)"', r"\1", sql)
    return sql.strip()


def test_normalise_sql():
    assert normalise_sql("""\
        CREATE TABLE "Node"( -- This is my table
            -- There are many like it but this one is mine
            A b, C D, "E F G", h)""") == \
        'CREATE TABLE Node(A b,C D,"E F G",h)'



from ..hookspecs import hookimpl
from datetime import datetime
import base64
import re
import hashlib
import httpx
import json
from ..zstd import get_decompressor, get_compressor
from urllib.parse import urlparse

FETCH_CACHE = 'fetch-cache'

_re = {}

def request_hash(url, request_headers):
    return base64.b85encode(hashlib.sha1((url + ':' + json.dumps(request_headers)).encode('utf-8')).digest()).decode()

@hookimpl(trylast=True)
def fetch_cached_url(conn, config, url, depth, request_headers):
    if not FETCH_CACHE in config or not config[FETCH_CACHE]:
        return

    max_age = None
    for option in config[FETCH_CACHE]:
        regex = option['url-regex']
        if regex in _re:
            compiled = _re[regex]
        else:
            compiled = re.compile(regex)
            _re[regex] = compiled

        if compiled.search(url) or compiled.search('depth:{}'.format(depth)):
            if max_age is None or max_age > option['max-age']:
                max_age = option['max-age']

    if max_age is None:
        return None

    hash = request_hash(url, request_headers)

    row = conn.execute("SELECT object, dict_id FROM dss_fetch_cache WHERE request_hash = ? AND fetched_at >= strftime('%Y-%m-%d %H:%M:%f', 'now', ?)", [hash, '-{} seconds'.format(max_age)]).fetchone()

    if not row:
        return None

    # TODO: consider batching these in a txn to reduce write lock contention
    with conn:
        conn.execute("UPDATE dss_fetch_cache SET read_at = strftime('%Y-%m-%d %H:%M:%f') WHERE request_hash = ?", [hash]).fetchone()


    object, dict_id = row

    object = get_decompressor(conn, dict_id).decompress(object)

    return json.loads(object.decode('utf-8'))

@hookimpl()
def after_fetch_url(conn, config, url, request_headers, response, fresh, fetch_duration):
    if not fresh:
        return

    # Don't store entries if we're disabled.
    if not FETCH_CACHE in config or not config[FETCH_CACHE]:
        return

    fetched_at = response['fetched_at']

    host = urlparse(url).hostname
    hash = request_hash(url, request_headers)
    response['_request_hash'] = hash

    dict_id = conn.execute('SELECT id FROM dss_zstd_dict WHERE host = ? AND active = 1 ORDER BY id DESC LIMIT 1', [host]).fetchone()

    if dict_id:
        dict_id = dict_id[0]

    compressor = get_compressor(conn, dict_id)

    object = json.dumps(response).encode('utf-8')
    object = compressor.compress(object)
    with conn:
        conn.execute('INSERT OR REPLACE INTO dss_fetch_cache(request_hash, host, url, fetched_at, read_at, object, dict_id) VALUES(?, ?, ?, ?, ?, ?, ?)', [hash, host, url, fetched_at, fetched_at, object, dict_id])


@hookimpl
def config_schema():
    from .. import ConfigSchema

    array_object = {
        'type': 'array',
        'items': {
            'type': 'object',
            'properties': {
                'url-regex': {
                    'type': 'string',
                    'title': 'URL Regex',
                },
                'max-age': {
                    'type': 'integer',
                    'title': "Max Age (seconds)"
                }
            },
            'required': ['url-regex', 'max-age']
        }
    }

    return ConfigSchema(
        schema = array_object,
        uischema = {
            "type": "Control",
            "scope": "#/properties/{}".format(FETCH_CACHE),
            "label": "Re-use previously downloaded pages"
        },

        key = FETCH_CACHE,
        group = 'Caching',
    )

@hookimpl
def config_default_value():
    return [
            {
                'url-regex': '.*',
                'max-age': 604800
            }
        ]



from datasette_scraper.plugins.discover_allow import canonicalize_url

def test_allow():
    assert canonicalize_url({}, 'from', 'to', 0) == None
    assert canonicalize_url({'discover-allow': []}, 'from', 'to', 0) == None

    assert canonicalize_url(
        {
            'discover-allow': [
                {
                    'from-url-regex': 'blah'
                }
            ]
        },
        'https://from.com/',
        'https://to.com/',
        0
    ) == False

    assert canonicalize_url(
        {
            'discover-allow': [
                {
                    'from-url-regex': 'from.com'
                }
            ]
        },
        'https://from.com/',
        'https://to.com/',
        0
    ) == True

    assert canonicalize_url(
        {
            'discover-allow': [
                {
                    'from-url-regex': 'from.com',
                    'depth': 3
                }
            ]
        },
        'https://from.com/',
        'https://to.com/',
        0
    ) == ('https://to.com/', 3)



import zstandard

DICT_SIZE = 112640
NUM_SAMPLES = 100

no_dict_compressor = zstandard.ZstdCompressor(level=9)
no_dict_decompressor = zstandard.ZstdDecompressor()

def get_dict(conn, dict_id):
    rv = conn.execute('SELECT dict FROM dss_zstd_dict WHERE id = ?', [dict_id]).fetchone()
    if not rv:
        raise Exception('unknown zstd dict_id {}'.format(dict_id))

    return zstandard.ZstdCompressionDict(rv[0])

def get_active_dict_id(conn, host):
    row = conn.execute('SELECT id FROM dss_zstd_dict WHERE active AND host = ? ORDER BY id DESC LIMIT 1', [host]).fetchone()

    if not row:
        return None

    dict_id, = row
    return dict_id

def get_compressor(conn, dict_id):
    if not dict_id:
        return no_dict_compressor

    return zstandard.ZstdCompressor(level=9, dict_data=get_dict(conn, dict_id))

def get_decompressor(conn, dict_id):
    if not dict_id:
        return no_dict_decompressor

    return zstandard.ZstdDecompressor(dict_data=get_dict(conn, dict_id))

def train_zstd_dict(conn, host):
    def get_samples(conn):
        hashes = conn.execute('SELECT object, dict_id FROM dss_fetch_cache WHERE host = ? ORDER BY RANDOM() LIMIT {}'.format(NUM_SAMPLES), [host]).fetchall()

        rv = []
        for (obj, dict_id) in hashes:
            decompressor = get_decompressor(conn, dict_id)
            rv.append(decompressor.decompress(obj))

        return rv

    samples = get_samples(conn)

    zdict = zstandard.train_dictionary(DICT_SIZE, samples)

    with conn:
        rv = conn.execute('INSERT INTO dss_zstd_dict(host, active, dict) VALUES (?, 1, ?)', [host, zdict.as_bytes()])
        return rv.lastrowid



from multiprocessing import Process, cpu_count
import sys
import os
import json
from datasette_scraper import ipc
from .config import enabled_databases
from .worker_ops import entrypoint_ops
from .worker_crawl import entrypoint_crawl

processes = []

NUM_WORKERS = cpu_count()

async def seed_crawl(db, job_id):
    await db.execute_write(
        'INSERT INTO dss_ops(type, config) VALUES (?, ?)',
        [ipc.SEED_CRAWL, json.dumps({'job-id': job_id})],
        block=True
    )

async def discover_missing_dictionaries(db):
    await db.execute_write(
        'INSERT INTO dss_ops(type) VALUES (?)',
        [ipc.DISCOVER_MISSING_DICTIONARIES],
        block=True
    )

def start_workers(datasette):
    # Don't start background workers if we're being tested under pytest.
    if "pytest" in sys.modules:
        return

    print('start_workers pid={}'.format(os.getpid()))
    db_map = {}
    for k, v in datasette.databases.items():
        if v.is_memory or not v.is_mutable:
            continue
        db_map[k] = v.path

    enabled_dbs = enabled_databases(datasette)

    if not enabled_dbs:
        raise Exception('datasette-scraper: not enabled in any databases, why are we starting workers?')

    p = Process(target=entrypoint_ops, args=(enabled_dbs, db_map), daemon=True)
    p.start()
    processes.append(('worker_ops', p))

    for i in range(NUM_WORKERS):
        p = Process(target=entrypoint_crawl, args=(enabled_dbs, db_map), daemon=True)
        p.start()
        processes.append(('worker_crawl', p))



import json
from datasette import Response
from datasette.utils import tilde_encode
from .config import enabled_databases
from .workers import seed_crawl, discover_missing_dictionaries
from .zstd import train_zstd_dict

async def crawl_exists(datasette, db, crawl_id):
    db = datasette.databases[db]
    rv = await db.execute('SELECT id FROM dss_crawl WHERE id = ?', [crawl_id])
    for row in rv:
        return True

    return False

def redirect_to_crawl(db_name, id):
    return Response.redirect('/{}/dss_crawl/{}'.format(db_name, id))

async def scraper_upsert(datasette, request):
    if request.method != 'POST':
        return Response('Unexpected method', status=405)

    form = await request.post_vars()

    id = int(form['id'])
    config = json.loads(form['config'])
    db_name = request.url_vars['db']

    name = config['name']
    config.pop('name')

    db = datasette.databases[db_name]

    if not id:
        rv = await db.execute_write('INSERT INTO dss_crawl(name, config) VALUES (?, ?)', [name, json.dumps(config)], block=True)
        id = rv.lastrowid
    else:
        await db.execute_write('UPDATE dss_crawl SET name = ?, config = ? WHERE id = ?', [name, json.dumps(config), id], block=True)

    return redirect_to_crawl(db_name, id)

async def scraper_discover_missing_dictionaries(datasette, request):
    if request.method != 'POST':
        return Response('Unexpected method', status=405)

    db_name = request.url_vars['db']
    db = datasette.databases[db_name]
    await discover_missing_dictionaries(db)

    return Response.redirect('/{}/dss_zstd_dict'.format(db.name))

async def scraper_host_rate_limit(datasette, request):
    if request.method != 'POST':
        return Response('Unexpected method', status=405)

    form = await request.post_vars()

    host = form['host']
    delay_seconds = float(form['delay_seconds'])

    db_name = request.url_vars['db']
    db = datasette.databases[db_name]

    await db.execute_write('UPDATE dss_host_rate_limit SET delay_seconds = ? WHERE host = ?', [delay_seconds, host], block=True)

    return Response.redirect('/{}/dss_host_rate_limit/{}'.format(db.name, tilde_encode(host)))


async def scraper_crawl_id(datasette, request):
    if request.method != 'GET':
        return Response('Unexpected method', status=405)

    id = int(request.url_vars["id"])
    db_name = request.url_vars['db']

    if not await crawl_exists(datasette, db_name, id):
        return Response('not found', status=404)

    context = {
        'dss_id': id
    }

    return Response.html(
        await datasette.render_template('/pages/-/scraper/crawl.html', context=context, request=request)
    )

async def scraper_crawl_id_start(datasette, request):
    if request.method != 'POST':
        return Response('Unexpected method', status=405)

    id = int(request.url_vars["id"])
    db_name = request.url_vars['db']

    if not await crawl_exists(datasette, db_name, id):
        return Response('not found', status=404)

    db = datasette.databases[db_name]

    rv = await db.execute_write("INSERT INTO dss_job(crawl_id) VALUES (?)", [id], block=True);
    job_id = rv.lastrowid

    await seed_crawl(db, job_id)

    return redirect_to_crawl(db_name, id)

async def scraper_crawl_id_cancel(datasette, request):
    if request.method != 'POST':
        return Response('Unexpected method', status=405)

    crawl_id = int(request.url_vars["id"])
    db_name = request.url_vars['db']

    if not await crawl_exists(datasette, db_name, crawl_id):
        return Response('not found', status=404)

    def cancel(conn):
        with conn:
            rv = conn.execute('SELECT id FROM dss_job WHERE crawl_id = ? AND finished_at IS NULL', [crawl_id])
            job_id, = rv.fetchone()

            if job_id:
                conn.execute("UPDATE dss_job SET status = 'cancelled', finished_at = strftime('%Y-%m-%d %H:%M:%f') WHERE id = ?", [job_id])
                conn.execute("DELETE FROM dss_crawl_queue WHERE job_id = ?", [job_id])

    db = datasette.databases[db_name]

    await db.execute_write_fn(cancel)

    return redirect_to_crawl(db_name, crawl_id)


async def scraper_crawl_id_edit(datasette, request):
    if request.method != 'GET':
        return Response('Unexpected method', status=405)

    db_name = request.url_vars['db']
    id = int(request.url_vars["id"])

    if not await crawl_exists(datasette, db_name, id):
        return Response('not found', status=404)

    return Response.html(
        await datasette.render_template('new-crawl.html', request=request)
    )

async def scraper_new(datasette, request):
    if request.method != 'GET':
        return Response('Unexpected method', status=405)

    return Response.html(
        await datasette.render_template('new-crawl.html', request=request)
    )


def get_routes(datasette):
    routes = []

    for db in enabled_databases(datasette):
        routes.append((r"^/(?P<db>{})/-/scraper/new$".format(db), scraper_new))
        routes.append((r"^/(?P<db>{})/-/scraper/upsert$".format(db), scraper_upsert))
        routes.append((r"^/(?P<db>{})/-/scraper/host-rate-limit$".format(db), scraper_host_rate_limit))
        routes.append((r"^/(?P<db>{})/-/scraper/crawl/(?P<id>[0-9]+)/start$".format(db), scraper_crawl_id_start))
        routes.append((r"^/(?P<db>{})/-/scraper/crawl/(?P<id>[0-9]+)/cancel$".format(db), scraper_crawl_id_cancel))
        routes.append((r"^/(?P<db>{})/-/scraper/discover-missing-dictionaries$".format(db), scraper_discover_missing_dictionaries))

    return routes



import pytest
from .utils import lazy_connection_factory, lazy_connection_factory_with_default
from .inserts import handle_insert

def test_lazy_connection_factory(tmp_path):
    db_name = tmp_path / "db.sqlite"
    raw_factory = lazy_connection_factory({ 'default': db_name })
    factory = lazy_connection_factory_with_default(raw_factory, 'default')

    conn = factory(None)
    conn2 = factory('default')

    assert conn == conn2
    with pytest.raises(Exception):
        factory('unknown')

def test_inserts(tmp_path):
    db_name = tmp_path / "db.sqlite"
    raw_factory = lazy_connection_factory({ 'default': db_name })
    factory = lazy_connection_factory_with_default(raw_factory, 'default')

    handle_insert(factory, { 'tab': [ { 'a!': 'A', 'b': 2} ] })

    # It should have created the table, with appropriate nullability
    conn = factory(None)
    rows = conn.execute('SELECT a, b FROM tab').fetchall()
    assert rows == [('A', 2)]

    handle_insert(factory, { 'tab': [ { 'a!': 'B', 'b': None } ] })

    # A row that already exists should be upserted
    handle_insert(factory, { 'tab': [ { 'a!': 'A', 'b': 123 } ] })

    rows = conn.execute("SELECT a, b FROM tab WHERE a = 'A'").fetchall()
    assert rows == [('A', 123)]

    # missing columns should be added
    handle_insert(factory, { 'tab': [ { 'a!': 'C', 'b': 123, 'c': 234 } ] })

    rows = conn.execute("SELECT * FROM tab WHERE a = 'C'").fetchall()
    assert rows == [('C', 123, 234)]

    # can specify dbname
    handle_insert(factory, { 'default': { 'tab2': [{'a!': 'a'}] } })
    rows = conn.execute('SELECT * FROM tab2').fetchall()
    assert rows == [('a',)]

    # an upsert of all pkeys should work
    handle_insert(factory, { 'default': { 'tab2': [{'a!': 'a'}] } })
    rows = conn.execute('SELECT * FROM tab2').fetchall()
    assert rows == [('a',)]

    # reserved words are OK, if perhaps a bit ill-advised
    handle_insert(factory, { 'from': [ {'from!': 'a' } ] })
    handle_insert(factory, { 'from': [ {'from!': 'a', 'insert': 'b' } ] })

    rows = conn.execute('SELECT * FROM "from"').fetchall()
    assert rows == [('a', 'b')]

def test_inserts_index(tmp_path):
    db_name = tmp_path / "db.sqlite"
    raw_factory = lazy_connection_factory({ 'default': db_name })
    factory = lazy_connection_factory_with_default(raw_factory, 'default')

    handle_insert(factory, { 'tab': [ {'a@': 'a', '__delete': True }, { 'a@': 'a', 'b': 1}, { 'a@': 'a', 'b': 2} ] })

    # It should have created the table, with appropriate nullability
    conn = factory(None)
    rows = conn.execute('SELECT a, b FROM tab').fetchall()
    assert rows == [('a', 1), ('a', 2)]

    handle_insert(factory, { 'tab': [ { 'a@': 'a', '__delete': True }, { 'a@': 'a', 'b': 3 } ] })
    rows = conn.execute('SELECT a, b FROM tab').fetchall()
    assert rows == [('a', 3)]

    # You can ask us to delete a row
    handle_insert(factory, { 'tab': [ { 'a@': 'a', '__delete': True} ] })

    rows = conn.execute("SELECT a, b FROM tab WHERE a = 'a'").fetchall()
    assert rows == []

    # reserved words are OK, if perhaps a bit ill-advised
    handle_insert(factory, { 'from': [ {'from@': 'a' } ] })
    handle_insert(factory, { 'from': [ {'from@': 'a', 'insert': 'b' }, { 'from@' : 'a', 'insert': 'b' } ] })

    rows = conn.execute('SELECT * FROM "from" ORDER BY 1,2').fetchall()
    assert rows == [('a', None), ('a', 'b'), ('a', 'b')]




from ..hookspecs import hookimpl
from urllib.parse import urlparse

MAX_PAGES_PER_DOMAIN = 'max-pages-per-domain'

@hookimpl
def canonicalize_url(conn, config, job_id, to_url):
    if MAX_PAGES_PER_DOMAIN in config:
        max_pages = config[MAX_PAGES_PER_DOMAIN]

        parsed = urlparse(to_url)
        host = parsed.hostname

        fetched, = conn.execute('SELECT COALESCE(SUM(fetched), 0) FROM dss_job_stats WHERE job_id = ? AND host = ?', [job_id, host]).fetchone()
        if fetched > max_pages:
            return 'max-pages-per-domain {} is less than fetched pages for domain {} ({})'.format(max_pages, host, fetched)

@hookimpl
def config_schema():
    from .. import ConfigSchema
    return ConfigSchema(
        schema = {
          'type': 'number',
        },
        uischema = {
            "type": "Control",
            "scope": '#/properties/{}'.format(MAX_PAGES_PER_DOMAIN)
        },
        key = MAX_PAGES_PER_DOMAIN,
        group = 'Links',
        sort = 120,
    )

@hookimpl
def config_default_value():
    return 1000



from setuptools import setup
import os

VERSION = "0.5.2"


def get_long_description():
    with open(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "README.md"),
        encoding="utf8",
    ) as fp:
        return fp.read()


setup(
    name="datasette-scraper",
    description="Adds website scraping abilities to Datasette.",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="Colin Dellow",
    url="https://github.com/cldellow/datasette-scraper",
    project_urls={
        "Issues": "https://github.com/cldellow/datasette-scraper/issues",
        "CI": "https://github.com/cldellow/datasette-scraper/actions",
        "Changelog": "https://github.com/cldellow/datasette-scraper/releases",
    },
    license="Apache License, Version 2.0",
    classifiers=[
        "Framework :: Datasette",
        "License :: OSI Approved :: Apache Software License"
    ],
    version=VERSION,
    packages=["datasette_scraper", "datasette_scraper.plugins"],
    entry_points={"datasette": ["scraper = datasette_scraper"]},
    install_requires=["datasette", "selectolax", "datasette-template-sql", "pluggy", "httpx", "zstandard", "more-itertools"],
    extras_require={"test": ["wheel", "pytest", "pytest-asyncio", "pytest-watch", "coverage"]},
    package_data={
        "datasette_scraper": ["static/*", "templates/*"]
    },
    python_requires=">=3.7",
)



import re
import json
from ..hookspecs import hookimpl
from ..utils import get_html_parser
from urllib.parse import urljoin

EXTRACT_JSON_LD_PRODUCT = 'extract-json-ld-product'

_re = {}

@hookimpl
def extract_from_response(config, url, response):
    if not EXTRACT_JSON_LD_PRODUCT in config:
        return {}

    rv = {}

    for options in config[EXTRACT_JSON_LD_PRODUCT]:
        if 'url-regex' in options:
            regex = options['url-regex']
            if regex in _re:
                compiled = _re[regex]
            else:
                compiled = re.compile(regex)
                _re[regex] = compiled

            if not compiled.search(url):
                continue

        dbname = options['database']
        tablename = options['table']

        if dbname in rv:
            database = rv[dbname]
        else:
            database = rv[dbname] = {}

        if tablename in database:
            table = database[tablename]
        else:
            table = database[tablename] = []

        parsed = get_html_parser(response)
        table.append({'source_url@': url, '__delete': True})
        for script in parsed.css('script[type="application/ld+json"]'):
            parsed = None
            try:
                parsed = json.loads(script.text())
            except:
                continue

            if not isinstance(parsed, dict):
                continue

            if not '@type' in parsed or parsed['@type'] != 'Product' or not '@context' in parsed or parsed['@context'] != 'http://schema.org/':
                continue

            products = extract_products(parsed)

            for product in products:
                product['source_url@'] = url

            table.extend(products)

    return rv

def extract_products(parsed):
    rv = []
    base = {}

    if 'name' in parsed:
        base['name'] = parsed['name']

    if 'sku' in parsed and parsed['sku'] and isinstance(parsed['sku'], str):
        base['sku'] = parsed['sku']

    if 'mpn' in parsed and parsed['mpn'] and isinstance(parsed['mpn'], str):
        base['mpn'] = parsed['mpn']

    if 'image' in parsed:
        if isinstance(parsed['image'], str):
            base['image'] = parsed['image']

        if isinstance(parsed['image'], list) and parsed['image']:
            base['image'] = parsed['image'][0]

    if 'id' in parsed:
        base['id'] = parsed['id']

    if 'url' in parsed:
        base['url'] = parsed['url']

    if 'brand' in parsed:
        if isinstance(parsed['brand'], dict) and 'name' in parsed['brand']:
            base['brand'] = parsed['brand']['name']

    if 'seller' in parsed:
        if isinstance(parsed['seller'], dict) and 'name' in parsed['seller']:
            base['seller'] = parsed['seller']['name']

    # This is a lot of boilerplate for not much use - maybe we make it configurable
    # to extract or not?
    #if 'description' in parsed:
    #    base['description'] = parsed['description']

    offers = []
    raw_offers = []
    if 'offers' in parsed and isinstance(parsed['offers'], dict):
        raw_offers = [parsed['offers']]
    elif 'offers' in parsed and isinstance(parsed['offers'], list):
        raw_offers = parsed['offers']

    valid_base = 'name' in parsed

    if not valid_base:
        return []

    for raw in raw_offers:
        new_dict = {}
        for k, v in base.items():
            new_dict[k] = v

        # https://schema.org/ItemAvailability
        if 'availability' in raw and raw['availability'].startswith('https://schema.org/'):
            new_dict['availability'] = raw['availability'][19:]

        # https://schema.org/ItemCondition
        if 'itemCondition' in raw and raw['itemCondition'].startswith('https://schema.org/'):
            new_dict['condition'] = raw['itemCondition'][19:]

        if 'sku' in raw:
            new_dict['sku'] = raw['sku']

        if 'url' in raw:
            new_dict['url'] = raw['url']

        if 'id' in raw:
            new_dict['id'] = raw['id']

        if 'priceSpecification' in raw:
            spec = raw['priceSpecification']

            if 'price' in spec:
                new_dict['price'] = spec['price']

            if 'priceCurrency' in spec:
                new_dict['currency'] = spec['priceCurrency']


        if 'price' in raw:
            new_dict['price'] = raw['price']

        if 'priceCurrency' in raw:
            new_dict['currency'] = raw['priceCurrency']

        rv.append(new_dict)

    if rv:
        return rv

    return [base]




@hookimpl
def config_schema():
    from .. import ConfigSchema
    return ConfigSchema(
        schema = {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'url-regex': {
                        'type': 'string',
                    },
                    'database': {
                        'type': 'string',
                    },
                    'table': {
                        'type': 'string',
                    },
                },
                'required': ['database', 'table']
            }
        },
        uischema = {
            "type": "Control",
            "scope": '#/properties/{}'.format(EXTRACT_JSON_LD_PRODUCT),
            'label': 'Extract JSON+LD product details'
        },
        key = EXTRACT_JSON_LD_PRODUCT,
        sort = 3000,
        group = 'Extracting',
    )

@hookimpl
def config_default_value():
    return []



import sqlite3
import itertools

def remove_sigil(key):
    if key[-1] == '!' or key[-1] == '@':
        return key[:-1]

    return key

def quote(k):
    return '"{}"'.format(k)

def remove_sigils(keys):
    return [remove_sigil(key) for key in keys]

def handle(counts, conn, dbname, tablename, obj):
    tpl = (dbname, tablename)

    inserted = 0
    updated = 0
    deleted = 0

    if '__delete' in obj:
        clause = {}
        for k, v in obj.items():
            if k == '__delete':
                continue

            clause[k] = v

        try:
            with conn:
                rv = conn.execute(
                    'DELETE FROM {} WHERE {}'.format(
                        quote(tablename),
                        ', '.join([quote(remove_sigil(k)) + ' = ?' for k in clause.keys()])
                    ),
                    list(clause.values())
                )
                deleted += rv.rowcount

            if tpl in counts:
                counts[tpl] = (counts[tpl][0] + inserted, counts[tpl][1] + updated, counts[tpl][2] + deleted)
            else:
                counts[tpl] = (inserted, updated, deleted)

            return
        except:
            # The delete might fail because the table hasn't been created yet; ignore it
            return


    args = [
        quote(tablename),
        ', '.join([quote(k) for k in remove_sigils(obj.keys())]),
        ', '.join(['?'] * len(obj.keys()))
    ]

    bindings = list(obj.values())
    stmt = 'INSERT INTO {} ({}) VALUES ({})'.format(*args)

    # If they've specified a pkey, do an UPSERT
    keys = obj.keys()
    pkeys = remove_sigils([pk for pk in keys if pk[-1] == '!'])

    if pkeys and len(pkeys) != len(keys):
        stmt += ' ON CONFLICT({}) DO UPDATE SET {}'.format(
            ', '.join([quote(k) for k in pkeys]),
            ', '.join(['{} = ?'.format(quote(remove_sigil(k))) for k in keys if k[-1] != '!'])
        )

        for k in keys:
            if k[-1] != '!':
                bindings.append(obj[k])
    elif pkeys:
        stmt += ' ON CONFLICT({}) DO NOTHING'.format(
            ', '.join([quote(k) for k in pkeys])
        )
    retry = True

    while retry:
        retry = False
        try:
            #print('stmt={}, values={}'.format(stmt, bindings))

            cur = conn.cursor()
            # This SELECT 1 is a little hacky - we need to force the cursor to show
            # us the lastrowid inserted on this connection.
            cur.execute('SELECT 1')
            lastrowid = cur.lastrowid
            cur.execute(stmt, bindings)
            if cur.lastrowid and cur.lastrowid != lastrowid:
                inserted += 1
            else:
                updated += 1
        except sqlite3.OperationalError as e:
            if e.args[0] == 'no such table: {}'.format(tablename):
                colspec = ',\n'.join(['  {} ANY {}'.format(quote(remove_sigil(k)), 'NOT NULL' if k[-1] == '!' else '')  for k in keys])

                pkeyspec = ''
                if pkeys:
                    pkeyspec = ', PRIMARY KEY({})'.format(', '.join([quote(k) for k in pkeys]))

                create_stmt = 'CREATE TABLE {} (\n{}{})'.format(quote(tablename), colspec, pkeyspec)
                #print(create_stmt)
                conn.execute(create_stmt)

                indexed_columns = [remove_sigil(k) for k in keys if k.endswith('@')]

                if indexed_columns:
                    create_index_stmt = 'CREATE INDEX {} ON {}({})'.format(quote("idx_" + tablename + "_multiple"), quote(tablename), ', '.join([quote(col) for col in indexed_columns]))
                    conn.execute(create_index_stmt)
                retry = True
            elif e.args[0].startswith('table {} has no column named '.format(tablename)):
                needs_column = e.args[0].split(' ')[-1]
                for k in keys:
                    if remove_sigil(k) == needs_column:
                        alter_stmt = 'ALTER TABLE {} ADD COLUMN {} ANY'.format(quote(tablename), quote(needs_column))
                        #print(alter_stmt)
                        conn.execute(alter_stmt)
                        retry = True
                        break

                if not retry:
                    raise e
            else:
                raise

    if tpl in counts:
        counts[tpl] = (counts[tpl][0] + inserted, counts[tpl][1] + updated, counts[tpl][2] + deleted)
    else:
        counts[tpl] = (inserted, updated, deleted)

def get_existing_rows(conn, table, obj, all_columns):
    clause = {}
    for k, v in obj.items():
        if k == '__delete':
            continue

        clause[k] = v
    stmt = 'SELECT {} FROM {} WHERE {}'.format(
        ', '.join([quote(remove_sigil(k)) + " AS " + quote(k) for k in all_columns.keys()]),
        quote(table),
        ' AND '.join(['{} = ?'.format(quote(remove_sigil(k))) for k in clause.keys()])
    )
    #print(stmt)
    all_hits = conn.execute(stmt, list(clause.values())).fetchall()
    rv = []
    for hit in all_hits:
        obj = {}
        for i, k in enumerate(all_columns.keys()):
            obj[k] = hit[i]
        rv.append(obj)
    return rv

def rows_match(python_rows, sqlite_rows):
    # There are some challenges here that mean a naive check for membership with `in`
    # will sometimes return a false negative. We will do the correct thing, but
    # fail to optimize.
    #
    # 1) We type our columns as ANY, so SQLite defaults to the NUMERIC affinity
    #    if the input can be coerced to a number. Thus we get 123, which won't
    #    match a Python string of '123' or '123.0'
    #
    # 2) Dicts that are passed to us can have fewer keys than columns returned,
    #    but as long as all the keys that are present are in the row from the
    #    db, that's fine.
    if len(python_rows) != len(sqlite_rows):
        return False

    banned = {}
    for python_row in python_rows:
        ok = False
        for i in range(len(sqlite_rows)):
            if i in banned:
                continue

            sqlite_row = sqlite_rows[i]

            if python_row == sqlite_row:
                banned[i] = True
                ok = True
                # delete the row to try to be a bit faster
                break

        if ok:
            continue

        # Look again, and be more forgiving about types and missing fields.
        for i in range(len(sqlite_rows)):
            if i in banned:
                continue

            sqlite_row = sqlite_rows[i]

            all_ok = True
            for k, python_value in python_row.items():
                sqlite_value = sqlite_row[k]

                if sqlite_value == python_value:
                    continue

                if isinstance(sqlite_value, int) or isinstance(sqlite_value, float):
                    if str(sqlite_value) == python_value:
                        continue

                    try:
                        if sqlite_value == float(python_value):
                            continue
                    except ValueError:
                        pass


                all_ok = False
                break

            if all_ok:
                banned[i] = True
                ok = True
                break

        if not ok:
            #print('failed to find python row {} in sqlite_rows={}'.format(python_row, sqlite_rows))
            return False

    return len(banned) == len(sqlite_rows)

def prune_raw(conn, dbname, table, rows):
    """Try to elide operations that would not actually change the DB."""

    # Ensure deletes always happen first; mutate the input so the slow path
    # has same behaviour as the fast path
    rows.sort(key=lambda x: not '__delete' in x)

    # NB: This doesn't try to be a general purpose solution!
    # It optimizes for a very narrow use case: when all of the rows being
    # touched have the same sigiled columns.

    # Case 1: pkey upserts - they all have the name! sigil
    # { 'name!': 'Colin', 'gender': 'm' }
    # { 'name!': 'Jenn', 'gender': 'f' }
    #
    # This represents a bunch of pkey upserts. A given row can be elided if
    # there is an existing row that has that subset of columns.

    # Case 2: indexed column or pkey deletes - they all have the url@ sigil
    # { 'url@': 'example.org', '__delete': True }
    #
    # This can be elided if no rows exist with the given url.

    # Case 3: indexed column upserts - they all have the url@ sigil
    # { 'url@': 'example.org', '__delete': True }
    # { 'url@': 'example.org', 'to': 'google.com' }
    # { 'url@': 'example.org', 'to': 'bing.com' }
    #
    # We only handle the happiest path here - if all rows are present in the db,
    # omit this sequence of instructions.

    # Any other case is ignored and falls through to the slow path.

    sigil_columns = None
    all_columns = {}
    for row in rows:
        x = {}
        for k in row.keys():
            if k[-1] == '!' or k[-1] == '@':
                x[k] = True
                all_columns[k] = k
            elif k != '__delete':
                all_columns[k] = k

        if sigil_columns == None:
            sigil_columns = x
        elif sigil_columns != x:
            print('! sigil mismatch: {} and  {}'.format(sigil_columns, x))
            return rows

    if not sigil_columns:
        print('! no sigils')
        return rows

    # They're all the same, fast path is permitted.
    #
    # Group the rows by the values of their sigils.
    key_columns = list(sigil_columns.keys())
    grouped_rows = itertools.groupby(rows, key=lambda row: tuple([row[k] for k in key_columns]))

    all_sigils_are_pkey = True
    all_sigils_are_index = True
    for sigil_column in sigil_columns:
        if sigil_column[-1] != '!':
            all_sigils_are_pkey = False
        if sigil_column[-1] != '@':
            all_sigils_are_index = False

    new_rows = []
    for (key, grouper) in grouped_rows:
        key_rows = list(grouper)

        if len(key_rows) == 1 and '__delete' in key_rows[0]:
            #print('optimizable delete: key={} rows={}'.format(key, key_rows[0]))
            #TODO: optimization: this could just be a SELECT EXISTS(...)
            existing_rows = get_existing_rows(conn, table, key_rows[0], all_columns)
            if existing_rows:
                new_rows.extend(key_rows)
        elif len(key_rows) > 1 and '__delete' in key_rows[0] and not '__delete' in key_rows[1]:
            #print('optimizable delete/insert: key={} rows={}'.format(key, key_rows))
            existing_rows = get_existing_rows(conn, table, key_rows[0], all_columns)
            #print(existing_rows)

            if not rows_match(key_rows[1:], existing_rows):
                #print('failed to optimize delete/insert')
                #print('optimizable delete/insert: key={} rows={}'.format(key, key_rows))
                new_rows.extend(key_rows)
        elif not '__delete' in key_rows[0] and all_sigils_are_pkey:
            #print('optimizable pkey: key={} rows={}'.format(key, key_rows))

            for row in key_rows:
                existing_rows = get_existing_rows(conn, table, row, all_columns)

                if len(existing_rows) != 1:
                    new_rows.append(row)
                elif existing_rows[0] != row:
                    new_rows.append(row)
        else:
            #print('not optimizable: key={} rows={}'.format(key, key_rows))
            new_rows.extend(key_rows)


    return new_rows

def prune(conn, dbname, table, rows):
    try:
        return prune_raw(conn, dbname, table, rows)
    except sqlite3.OperationalError as e:
        # Probably this table has not been created yet, or the user is adding
        # a column. Fall through to slow path.
        #print(e)
        return rows

def handle_insert(factory, insert):
    # This function does a _lot_. It can:
    # - create tables
    # - create indexes
    # - add columns to existing tables
    # - delete rows that match a predicate
    # - insert rows
    #
    # Additionally, it tries to avoid doing write transactions needlessly. Given a set of input,
    # it will try to determine if the database would actually be changed. It does this interrogation
    # using only read transactions. If the database would not be changed, no write lock is required.
    #
    # If the database _would_ be changed, it tries to issue the smallest set of changes, so that
    # other writers are blocked for the smallest time and we don't bloat the wal journal.
    counts = {}

    for k, v in insert.items():
        dbname = None

        if isinstance(v, dict):
            dbname = k
            conn = factory(dbname)

            for table, objs in v.items():
                objs = prune(conn, dbname, table, objs)
                counts[(dbname, table)] = (0, 0, 0)
                for obj in objs:
                    handle(counts, conn, dbname, table, obj)
        else:
            conn = factory(dbname)
            table = k
            objs = v
            counts[(dbname, table)] = (0, 0, 0)
            objs = prune(conn, dbname, table, objs)
            for obj in objs:
                handle(counts, conn, dbname, table, obj)

    return counts



from ..hookspecs import hookimpl
from urllib.parse import urlparse

FOLLOW_REDIRECTS = 'follow-redirects'

@hookimpl
def discover_urls(config, url, response):
    if not FOLLOW_REDIRECTS in config or not config[FOLLOW_REDIRECTS]:
        return []

    for (k, v) in response['headers']:
        if k == 'location':
            return [v]

    return []


@hookimpl
def config_schema():
    from .. import ConfigSchema
    return ConfigSchema(
        schema = {
          'type': 'boolean',
          'title': 'Follow HTTP 3xx redirects',
        },
        uischema = {
            "type": "Control",
            "scope": '#/properties/{}'.format(FOLLOW_REDIRECTS)
        },
        key = FOLLOW_REDIRECTS,
        group = 'Links',
        sort = -1
    )

@hookimpl
def config_default_value():
    return True
